import abc
import dataclasses
import enum
import re
import typing as t
from pathlib import Path

from nr.util.functional import Consumer
from nr.util.generic import T
from nr.util.git import Git as _Git, NoCurrentBranchError


class FileStatus(enum.Enum):
    NONE = enum.auto()
    ADDED = enum.auto()
    MODIFIED = enum.auto()
    RENAMED = enum.auto()
    DELETED = enum.auto()
    UNKNOWN = enum.auto()


@dataclasses.dataclass
class FileInfo:
    path: Path
    staging: FileStatus
    disk: FileStatus


@dataclasses.dataclass
class Remote:
    name: str
    url: str
    default: bool


@dataclasses.dataclass
class Author:
    name: str | None
    email: str | None


class Vcs(abc.ABC):
    """Interface to perform actions on a local version control system and its remote counterpart."""

    @abc.abstractmethod
    def get_toplevel(self) -> Path:
        ...

    @abc.abstractmethod
    def get_web_url(self) -> str | None:
        """Try to identify the web URL of the repository."""

    @abc.abstractmethod
    def get_remotes(self) -> t.Sequence[Remote]:
        """Return a sequence of the known remote copies that the local version can be synced with. At least one of the
        returned remotes should be marked as the default remote via #Remote.default."""

    @abc.abstractmethod
    def get_current_branch(self) -> str | None:
        """Return the name of the current branch in the local repository. This is used by Slap to ensure that the user
        is on the right branch before creating a new release."""

    @abc.abstractmethod
    def get_author(self) -> Author:
        """Determine the author details from the repository or VCS settings."""

    @abc.abstractmethod
    def get_all_files(self) -> t.Sequence[Path]:
        """Return a sequence of all the files known to the VCS."""

    @abc.abstractmethod
    def get_changed_files(self) -> t.Sequence[FileInfo]:
        """Return the status of all files in the version control repository that have been changed or are unknown to the
        VCS (and not ignored)."""

    @abc.abstractmethod
    def get_file_contents(self, file: Path, revision: str) -> bytes | None:
        """Return the contents of the file in a given revision. Return `None` if the file does not exist."""

    @abc.abstractmethod
    def commit_files(
        self,
        files: t.Sequence[Path],
        commit_message: str,
        *,
        tag_name: str | None = None,
        push: Remote | None = None,
        force: bool = False,
        allow_empty: bool = False,
        email: str | None = None,
        name: str | None = None,
        log_line: Consumer[str] | None = None,
    ) -> None:
        """Commit the given files into the VCS, and optionally create a tag with the given name. If a remote is
        specified for the *push* argument, the commit that was just created on the current branch as well as the tag
        name if one was specified will be pushed to the remote."""

    @classmethod
    @abc.abstractclassmethod
    def detect(cls: type[T], path: Path) -> T | None:
        ...


class Git(Vcs):
    def __init__(self, directory: Path) -> None:
        self._git = _Git(directory)
        assert self._git.get_toplevel() is not None, f"Not a Git repository: {directory}"

    def __repr__(self) -> str:
        return f'Git("{self._git.path}")'

    def get_toplevel(self) -> Path:
        toplevel = self._git.get_toplevel()
        assert toplevel is not None
        return Path(toplevel)

    def get_web_url(self) -> str | None:
        remote = next((r for r in self._git.remotes() if r.name == "origin"), None)
        if not remote:
            return None
        url = remote.fetch
        if url.endswith(".git"):
            url = url[:-4]
        if url.startswith("http"):
            return url
        match = re.match(r"\w+@(.*)$", url)
        if match:
            return "https://" + match.group(1)
        return None

    def get_remotes(self) -> t.Sequence[Remote]:
        result = []
        for remote in self._git.remotes():
            default = remote.name == "origin"
            result.append(Remote(remote.name, remote.push, default))
        return result

    def get_current_branch(self) -> str | None:
        try:
            return self._git.get_current_branch_name()
        except NoCurrentBranchError:
            return None

    def get_author(self) -> Author:
        return get_git_author(self._git.path)

    def get_all_files(self) -> t.Sequence[Path]:
        return [self._git.path / f for f in self._git.get_files()]

    def get_changed_files(self) -> t.Sequence[FileInfo]:
        result = []
        for file in self._git.get_status():
            result.append(
                FileInfo(Path(file.filename), self._git_file_status(file.mode[0]), self._git_file_status(file.mode[1]))
            )
        return result

    def get_file_contents(self, file: Path, revision: str) -> bytes | None:
        try:
            return self._git.get_file_contents(str(file), revision)
        except FileNotFoundError:
            return None

    def commit_files(
        self,
        files: t.Sequence[Path],
        commit_message: str,
        *,
        tag_name: str | None = None,
        push: Remote | None = None,
        force: bool = False,
        allow_empty: bool = False,
        email: str | None = None,
        name: str | None = None,
        log_line: Consumer[str] | None = None,
    ) -> None:
        # TODO (@NiklasRosenstein): Capture stdout from Git subprocess and redirect to log_line().
        self._git.add([str(f.resolve()) for f in files])

        commit_command = ["git"]
        if email:
            commit_command += ["-c", f"user.email={email}"]
        if name:
            commit_command += ["-c", f"user.name={name}"]
        commit_command += ["commit", "-m", commit_message]
        if allow_empty:
            commit_command += ["--allow-empty"]
        self._git.check_call(commit_command)

        if tag_name is not None:
            self._git.tag(tag_name, force)
        if push is not None:
            refs = [self._git.get_current_branch_name()]
            if tag_name is not None:
                refs.append(tag_name)
            self._git.push(push.name, *refs, force=force)

    @classmethod
    def detect(cls, path: Path) -> t.Union["Git", None]:
        if _Git(path).get_toplevel() is not None:
            return Git(path)
        return None

    @staticmethod
    def _git_file_status(mode: str) -> FileStatus:
        return {
            " ": FileStatus.NONE,
            "A": FileStatus.ADDED,
            "M": FileStatus.MODIFIED,
            "R": FileStatus.RENAMED,
            "D": FileStatus.DELETED,
            "?": FileStatus.UNKNOWN,
        }[mode]


def get_git_author(path: Path | None = None) -> Author:
    import subprocess as sp

    git = _Git(path)
    try:
        name = git.get_config("user.name") or git.get_config("user.name", global_=True)
        email = git.get_config("user.email") or git.get_config("user.email", global_=True)
    except sp.CalledProcessError as exc:
        if exc.returncode != 1:
            raise
        return Author(None, None)
    return Author(name, email)


def detect_vcs(path: Path) -> Vcs | None:
    for cls in [Git]:
        if vcs := cls.detect(path):
            return vcs
    return None



from __future__ import annotations

import dataclasses
import logging
import os
import typing as t
from pathlib import Path

import typing_extensions as te
from databind.core.settings import Alias, ExtraKeys

from slap.application import Application, Command, option
from slap.configuration import Configuration
from slap.ext.application.venv import VenvAwareCommand
from slap.plugins import ApplicationPlugin
from slap.project import Project

if t.TYPE_CHECKING:
    from slap.install.installer import Indexes
    from slap.python.dependency import Dependency
    from slap.python.environment import PythonEnvironment


logger = logging.getLogger(__name__)

python_option = option(
    "--python",
    "-p",
    description="The Python executable to install to.",
    flag=False,
)


@t.overload
def get_active_python_bin(cmd: Command) -> str:
    ...


@t.overload
def get_active_python_bin(cmd: Command, fallback: te.Literal[False]) -> str | None:
    ...


def get_active_python_bin(cmd: Command, fallback: bool = True) -> str | None:
    """Returns the active Python installation."""

    if hasattr(cmd, "_python_bin"):
        python = cmd._python_bin  # type: ignore
    else:
        python = cmd.option("python")
        if not python:
            python = os.getenv("PYTHON")
    if fallback:
        python = python or "python"
    cmd._python_bin = python
    return python


def venv_check(cmd: Command, message="refusing to install", env: PythonEnvironment | None = None) -> bool:
    from slap.python.environment import PythonEnvironment

    if not cmd.option("no-venv-check"):
        env = env or PythonEnvironment.of(get_active_python_bin(cmd))
        if not env.is_venv():
            cmd.line_error(f"error: {message} because you are not in a virtual environment", "error")
            cmd.line_error("       enter a virtual environment or use <opt>--no-venv-check</opt>", "error")
            cmd.line_error(f"       the Python executable you are targeting is <s>{env.executable}</s>", "error")
            return False
    return True


@dataclasses.dataclass
@ExtraKeys(True)
class InstallConfig:
    """Separate install configuration under Slap that is entirely separate from the build system that is being used.
    This configures the behaviour of the `slap install` command."""

    #: Additional extras that are installable with `--extras` or `--only-extras`. These are taken into account
    #: for projects as well as the mono-repository root. They may contain semantic version selector (same style
    #: as supported by Poetry).
    extras: dict[str, list[str]] = dataclasses.field(default_factory=dict)

    #: A list of extra names to install in addition to `dev` when using `slap install` (unless `--no-dev` is
    #: specified). If this option is not set, _all_ extras are installed.
    dev_extras: t.Annotated[list[str] | None, Alias("dev-extras")] = None


class InstallCommandPlugin(VenvAwareCommand, ApplicationPlugin):
    """Install your project and its dependencies via Pip."""

    app: Application
    name = "install"
    options = VenvAwareCommand.options + [
        option(
            "--only",
            description="Path to the subproject to install only. May still cause other projects to be installed if "
            "required by the selected project via inter dependencies, but only their run dependencies will be "
            "installed.",
            flag=False,
        ),
        option(
            "--link",
            description="Symlink the root project using <opt>slap link</opt> instead of installing it directly.",
        ),
        option(
            "--no-dev",
            description="Do not install development dependencies.",
        ),
        option(
            "--no-root",
            description="Do not install the package itself, but only its dependencies.",
        ),
        option(
            "--extras",
            description='A comma-separated list of extras to install. Note that <s>"dev"</s> is a valid extras.',
            flag=False,
        ),
        option(
            "--only-extras",
            description='Install only the specified extras. Note that <s>"dev"</s> is a valid extras.',
            flag=False,
        ),
        option(
            "--upgrade",
            description="Upgrade already installed packages.",
            flag=True,
        ),
        option(
            "--from",
            description="Install another Slap project from the given directory.",
            flag=False,
        ),
        option(
            "--index",
            description="Set an index URL to install from. Must be formatted like <s>name=myindex,url=https://...</s> "
            "The username and password may be specified as arguments as well. Note that this format requires "
            "that commas are urlencoded if they are presented in the URL. This argument can be used to override the "
            "URL of the index that a dependency is installed from (the dependency must specify a `source` in "
            "`pyproject.toml`) and/or inject credentials.",
            flag=False,
            multiple=True,
        ),
        option(
            "--extra-index",
            description="Deprecated. Same as --index.",
            flag=False,
            multiple=True,
        ),
        python_option,
    ]

    def load_configuration(self, app: Application) -> None:
        from databind.json import load

        self.config: dict[Configuration, InstallConfig] = {}
        for obj in app.configurations():
            self.config[obj] = load(obj.raw_config().get("install", {}), InstallConfig, filename=str(obj))
        return None

    def activate(self, app: Application, config: None) -> None:
        self.app = app
        app.cleo.add(self)

    def handle(self) -> int:
        """
        Installs the requirements of the package using Pip.
        """

        from nr.util.stream import Stream

        from slap.install.installer import InstallOptions, PipInstaller, get_indexes_for_projects
        from slap.python.dependency import PathDependency, PypiDependency, parse_dependencies
        from slap.python.environment import PythonEnvironment

        if not self._validate_args():
            return 1

        result = super().handle()
        if result != 0:
            return result

        if from_dir := self.option("from"):
            # TODO (@NiklasRosenstein): This is pretty hacky, but it works so far. We basically modify the application
            #       object as if it was running from the specified path all along.
            self.app.__init__(Path(from_dir).absolute())  # type: ignore
            self.app.load_plugins()
            self.load_configuration(self.app)

        python_environment = PythonEnvironment.of(get_active_python_bin(self))
        if not venv_check(self, env=python_environment):
            return 1

        projects = self._get_projects_to_install()
        if not projects:
            return 1

        # Get a list of the projects that need to be installed that also includes all the projects required through
        # interdependencies between the projects.
        projects_plus_dependencies = (
            Stream(projects)
            .map(lambda p: p.get_interdependencies(self.app.repository.projects(), recursive=True))
            .concat()
            .append(projects)
            .collect()
        )

        install_extras = self._get_extras_to_install()
        discovered_extras = {"dev"}  # Not discovering a 'dev' extra should not trigger a warning
        dependencies: list[Dependency] = []

        # Collect the run dependencies to install.
        for project in projects_plus_dependencies:
            assert project.is_python_project, "Project.is_python_project is deprecated and expected to always be true"
            deps = project.dependencies()

            if (
                not self.option("no-root")
                and not self.option("link")
                and not self.option("only-extras")
                and project.packages()
            ):
                # Install the project itself directory unless certain flags turn this behavior off.
                dependencies.append(PathDependency(project.dist_name() or project.id, project.directory))

            elif not self.option("only-extras"):
                # Install the run dependencies of the project.
                dependencies += deps.run

        # Collect dev dependencies and extras from the project.
        for project in projects:
            deps = project.dependencies()

            if (not self.option("no-dev") and not self.option("only-extras")) or "dev" in install_extras:
                # Install the development dependencies of the project.
                dependencies += deps.dev

            # Determine the extras to install for the current project. This changes on development installs because
            # we always consider the ones configured in #InstallConfig.dev_extras.
            current_project_install_extras = set(install_extras)
            if not self.option("no-dev"):
                config = self.config[project]
                if config.dev_extras is not None:
                    current_project_install_extras.update(config.dev_extras)

            # Append the extra dependencies from the project. We ignore 'dev' here because we already took care of
            # deciding when to install dev dependencies.
            for extra in current_project_install_extras:
                extra_deps = deps.extra.get(extra)
                if extra_deps is not None:
                    discovered_extras.add(extra)
                    dependencies += extra_deps

        # Look for extras also in the Slap specific install configuration.
        for _, config in self.config.items():
            for extra in install_extras:
                dependencies += parse_dependencies(config.extras.get(extra, []))
                discovered_extras.add(extra)

        if missing_extras := install_extras - discovered_extras:
            self.line_error(f"error: extras that do not exist: <fg=yellow>{missing_extras}</fg>", "error")
            return 1

        # Remove dependencies that reference projects in the same mono repository, as we are installing them separately.
        project_names = {project.dist_name() for project in projects_plus_dependencies}
        dependencies = [
            dependency
            for dependency in dependencies
            if not (isinstance(dependency, PypiDependency) and dependency.name in project_names)
        ]

        if not dependencies:
            self.line("nothing to install.", "info")
            return 0

        options = InstallOptions(
            indexes=get_indexes_for_projects(projects),
            quiet=self.option("quiet"),
            upgrade=self.option("upgrade"),
        )
        self._update_indexes_from_cli(options.indexes)

        installer = PipInstaller(self)
        status_code = installer.install(dependencies, python_environment, options)
        if status_code != 0:
            return status_code

        if self.option("link"):
            self.link_project(Path(self.option("from") or "."))

        return 0

    def _validate_args(self) -> bool:
        """Validate combinations of command-line args and options."""

        for a, b in [("only-extras", "extras"), ("no-root", "link"), ("only-extras", "link")]:
            if self.option(a) and self.option(b):
                self.line_error(f"error: conflicting options <opt>--{a}</opt> and <opt>--{b}</opt>", "error")
                return False

        return True

    def _get_projects_to_install(self) -> list[Project]:
        """Return the list of Slap projects to install."""

        if only_project := self.option("only"):
            project_path = Path(only_project).resolve()
            projects = [p for p in self.app.repository.projects() if p.directory.resolve() == project_path]
            if not projects:
                self.line_error(f'error: "{only_project}" does not point to a project', "error")
                return []
            assert len(projects) == 1, projects
            return projects

        else:
            if not self.app.repository.projects:
                self.line_error("error: no projects found")
                return []
            return self.app.repository.get_projects_ordered()

    def _get_extras_to_install(self) -> set[str]:
        """Return a set of the extras that should be installed."""

        extras = set(map(str.strip, (self.option("extras") or self.option("only-extras") or "").split(",")))
        extras.discard("")

        if not self.option("no-dev") and not self.option("only-extras"):
            extras.add("dev")

        if not self.option("no-dev") and self.app.repository in self.config:
            # Add the dev extras from the repository configuration.
            extras.update(self.config[self.app.repository].dev_extras or [])

        return extras

    def _update_indexes_from_cli(self, indexes: Indexes) -> None:
        from slap.install.installer import IndexSpec

        for extra in (*self.option("extra-index"), *self.option("index")):
            spec = IndexSpec.parse(extra)
            if spec.name in indexes.urls:
                spec.url = spec.url or indexes.urls[spec.name]
            else:
                logger.warning('passed an --index option for a source that does not exist (source: "%s")', spec.name)
            indexes.urls[spec.name] = spec.url_with_auth

    # SymlinkHelper

    def get_dependencies_for_project(self, project: Path) -> list[Dependency]:
        # TODO (@NiklasRosenstein): Implement this method
        raise NotImplementedError

    def link_project(self, project: Path) -> None:
        from slap.ext.application.link import link_repository

        app = Application(project)
        link_repository(self.io, app.repository, python=get_active_python_bin(self))



import typing as t

from cleo.commands.help_command import HelpCommand as _HelpCommand  # type: ignore[import]
from cleo.commands.list_command import ListCommand  # type: ignore[import]
from cleo.formatters.formatter import Formatter  # type: ignore[import]
from cleo.formatters.style import Style  # type: ignore[import]
from cleo.io.io import IO  # type: ignore[import]


@t.overload
def add_style(
    io: IO | Formatter,
    name: str,
    foreground: str | None = ...,
    background: str | None = ...,
    options: list[str] | None = ...,
) -> None:
    ...


@t.overload
def add_style(
    io: IO | Formatter,
    name: str,
    style: Style,
) -> None:
    ...


def add_style(  # type: ignore[misc]
    io: IO | Formatter,
    name: str,
    foreground: str | Style | None = None,
    background: str | None = None,
    options: list[str] | None = None,
    *,
    style: Style | None = None,
) -> None:
    """
    Add a style to a Cleo IO or Formatter instance.
    """

    if style is not None:
        assert foreground is None and background is None and options is None
    elif isinstance(foreground, Style):
        style = foreground
        assert background is None and options is None
    else:
        style = Style(foreground, background, options)

    if isinstance(io, IO):
        io.output.formatter.set_style(name, style)
        io.error_output.formatter.set_style(name, style)
    elif isinstance(io, Formatter):
        io.set_style(name, style)
    else:
        raise TypeError(f"expected IO|Formatter, got {type(io).__name__}")


class HelpCommand(_HelpCommand, ListCommand):

    arguments = ListCommand.arguments

    def handle(self) -> int:
        self.io.input._arguments["namespace"] = None
        if not self._command:
            return ListCommand.handle(self)
        return _HelpCommand.handle(self)



import typing as t

from poetry.core.semver.version import Version  # type: ignore[import]

from slap.plugins import VersionIncrementingRulePlugin


def incrementing_rule(func: t.Callable[[Version], Version]) -> type[VersionIncrementingRulePlugin]:
    class _Incrementor(VersionIncrementingRulePlugin):
        def increment_version(self, version: Version) -> Version:
            return func(version)

    _Incrementor.__name__ = func.__name__
    return _Incrementor


@incrementing_rule
def major(version: Version) -> Version:
    return version.next_major()


@incrementing_rule
def premajor(version: Version) -> Version:
    return version.next_major().first_prerelease()


@incrementing_rule
def minor(version: Version) -> Version:
    return version.next_minor()


@incrementing_rule
def preminor(version: Version) -> Version:
    return version.next_minor().first_prerelease()


@incrementing_rule
def patch(version: Version) -> Version:
    return version.next_patch()


@incrementing_rule
def prepatch(version: Version) -> Version:
    return version.next_patch().first_prerelease()


@incrementing_rule
def prerelease(version: Version) -> Version:
    if version.is_unstable():
        assert version.pre
        return Version(version.epoch, version.release, version.pre.next())
    else:
        return version.next_patch().first_prerelease()



""" Implements the default package detection plugin. """

from __future__ import annotations

import abc
import re
import typing as t
from pathlib import Path

from nr.util.algorithm.longest_common_substring import longest_common_substring
from nr.util.fs import get_file_in_directory
from setuptools import find_namespace_packages, find_packages

from slap.plugins import ProjectHandlerPlugin
from slap.project import Package, Project
from slap.release import VersionRef, match_version_ref_pattern, match_version_ref_pattern_on_lines

if t.TYPE_CHECKING:
    from slap.python.dependency import Dependency

IGNORED_MODULES = ["test", "tests", "docs", "build"]


def detect_packages(directory: Path) -> list[Package]:
    """Detects the Python packages in *directory*, making an effort to identify namespace packages correctly."""

    if not directory.is_dir():
        return []

    assert isinstance(directory, Path)
    modules = list(set(find_namespace_packages(str(directory)) + find_packages(str(directory))))

    # Also support toplevel modules.
    for path in directory.iterdir():
        if path.is_file() and path.suffix == ".py" and path.stem not in modules:
            modules.append(path.stem)

    # Remove modules that seem to be other Python projects.
    modules = [m for m in modules if not (directory / m.partition(".")[0] / "pyproject.toml").is_file()]

    if not modules:
        return []

    paths = {}
    for module in modules:
        tlm_file = directory / (module + ".py")
        pkg_file = directory / Path(*module.split("."), "__init__.py")
        use_file = tlm_file if tlm_file.is_file() else pkg_file.parent if pkg_file.is_file() else None
        if use_file is not None:
            paths[module] = use_file

    modules = [m for m in modules if m in paths]

    modules = [
        m for m in modules if m not in IGNORED_MODULES and ("." not in m or m.split(".")[0] not in IGNORED_MODULES)
    ]

    if len(modules) > 1:
        # If we stil have multiple modules, we try to find the longest common path.
        common = longest_common_substring(*(x.split(".") for x in modules), start_only=True)
        if not common:
            return []
        modules = [".".join(common)]

    return [Package(module, paths[module], directory) for module in modules]


class BaseProjectHandler(ProjectHandlerPlugin):
    """Base class for other project handlers. It cannot be used directly by a project."""

    package_dirs: t.Sequence[str] = ("src", ".")

    def __repr__(self) -> str:
        return type(self).__name__

    def get_readme(self, project: Project) -> str | None:
        path = get_file_in_directory(project.directory, "README", ["README.rst"])
        return path.name if path else None

    def get_packages(self, project: Project) -> list[Package] | None:
        """Detects packages in #package_dirs."""

        source_dir = project.config().source_directory
        if source_dir:
            return detect_packages(project.directory / source_dir)
        else:
            for source_dir in self.package_dirs:
                packages = detect_packages(project.directory / source_dir)
                if packages:
                    return packages
        return []


class PyprojectHandler(BaseProjectHandler):
    """A subclass that implements some functionality based on whether the project is configured using a
    `pyproject.toml` file."""

    def get_version_refs(self, project: Project) -> list[VersionRef]:
        """Returns the version ref in `pyproject.toml` it can be found, as well as the version references of project
        interdependencies (you can disable the interdependencies bit by setting `tool.slap.release.interdependencies`
        setting to `False` on the Slap root directory, usually in a `slap.toml` file)."""

        PYPROJECT_TOML_PATTERN = r'^version\s*=\s*[\'"]?(.*?)[\'"]'
        version_ref = match_version_ref_pattern(project.pyproject_toml.path, PYPROJECT_TOML_PATTERN, None)
        refs = [version_ref] if version_ref else []
        if interdependencies_enabled(project):
            refs += get_pyproject_interdependency_version_refs(project)
        return refs

    def add_dependency(self, project: Project, dependency: Dependency, where: str) -> None:
        """Adds a dependency to the respective location in the `pyproject.toml` file."""

        import tomlkit
        import tomlkit.container
        import tomlkit.items

        root = tomlkit.parse(project.pyproject_toml.path.read_text())
        keys, value = self.get_add_dependency_toml_location_and_config(project, dependency, where)
        assert isinstance(value, list | dict), type(value)

        current: tomlkit.items.Item | tomlkit.container.Container = root
        for idx, key in enumerate(keys):
            # NOTE (@NiklasRosenstein): I have no clue why this is needed, but if we don't access the internal
            #   container of this "OutOfOrderTableProxy", it appears the mutation later on doesn't actually work.
            if isinstance(current, tomlkit.container.OutOfOrderTableProxy):
                current = current._internal_container

            if not isinstance(current, tomlkit.items.Table | tomlkit.container.Container):
                break  # Will triger an error below

            if key not in current:
                if idx == len(keys) - 1:
                    current[key] = tomlkit.array() if isinstance(value, list) else tomlkit.table()
                else:
                    current[key] = tomlkit.table()

            current = current[key]

        if isinstance(value, list):
            if not isinstance(current, tomlkit.items.Array):
                raise RuntimeError(f"expected array at {keys!r}, got {type(current).__name__}")
            for v in value:
                current.append(v)
        elif isinstance(value, dict):
            if not isinstance(current, tomlkit.items.Table | tomlkit.container.Container):
                raise RuntimeError(f"expected table at {keys!r}, got {type(current).__name__}")
            current.update(value)
        else:
            assert False, type(value)

        project.pyproject_toml.path.write_text(tomlkit.dumps(root))

    @abc.abstractmethod
    def get_add_dependency_toml_location_and_config(
        self,
        project: Project,
        dependency: Dependency,
        where: str,
    ) -> tuple[list[str], list | dict]:
        """Return the location and configuration to inject/update the TOML project configuration with in order
        to add the given *dependency* to the project.

        Returns:
          1. The sequence of keys that contains the dependencies based on the *where* argument.
          2. The element to inject extend/merge into the key at the location specified by the first tuple element.
        """


def interdependencies_enabled(project: Project) -> bool:
    return bool(project.repository.raw_config().get("release", {}).get("interdependencies", True))


def get_pyproject_interdependency_version_refs(project: Project) -> list[VersionRef]:
    """Identifies version references of another project in the set of projects loaded in the application. This is
    relevant in case when Slap is used in a mono-repository where all projects share the same version, and bumping
    version numbers should also bump the version number of dependencies between projects in that mono-repository."""

    pyproject_file = project.pyproject_toml.path
    other_projects: list[str] = [
        t.cast(str, p.dist_name())
        for p in project.repository.projects()
        if p.is_python_project and p is not project and p.dist_name()
    ]

    refs = []

    SELECTOR = r"([\^<>=!~\*]*)(?P<version>\d+\.[\w\d\.\-]+)"

    for name in other_projects:
        # Look for something that looks like a version number. In common TOML formats, that is usually as an entire
        # requirement string or as an assignment.
        expressions = [
            # This first one matches TOML key/value pairs.
            r'([\'"])?' + re.escape(name) + r'\1\s*=\s*([\'"])' + SELECTOR + r"\1",
            re.escape(name) + r'\s*=\s*([\'"])' + SELECTOR + r"\1",
            # This second one matches a TOML string that contains the dependency.
            r'([\'"])' + re.escape(name) + r"(?![^\w\d\_\.\-\ ])\s*" + SELECTOR + r"\1\s*($|,|\]|\})",
        ]

        for expr in expressions:
            refs += match_version_ref_pattern_on_lines(pyproject_file, expr)

    return refs



""" Commands that produce reports. """

import json
import logging
import typing as t

import pkg_resources

from slap.application import Application, option
from slap.ext.application.venv import VenvAwareCommand
from slap.plugins import ApplicationPlugin

if t.TYPE_CHECKING:
    from slap.python.dependency import Dependency

logger = logging.getLogger(__name__)


class ReportDependenciesCommand(VenvAwareCommand):
    """Reports the installed run dependencies of your current project(s) as JSON."""

    name = "report dependencies"
    options = VenvAwareCommand.options + [
        option(
            "extras",
            description="A comma-separated list of extra dependencies to include.",
            flag=False,
        ),
        option("with-license-text", description="Include license text in the output."),
    ]

    def __init__(self, app: Application) -> None:
        super().__init__()
        self.app = app

    def handle(self) -> int:
        import databind.json
        import tqdm  # type: ignore[import]

        from slap.python.environment import DistributionGraph, PythonEnvironment, build_distribution_graph
        from slap.python.pep508 import filter_dependencies

        result = super().handle()
        if result != 0:
            return result

        extras = set(filter(bool, map(str.strip, (self.option("extras") or "").split(","))))

        requirements: list[Dependency] = []
        for project in self.app.repository.projects():
            requirements += project.dependencies().run
            if "dev" in extras:
                requirements += project.dependencies().dev
            for extra in extras:
                requirements += project.dependencies().extra.get(extra, [])

        dists_cache: dict[str, pkg_resources.Distribution | None] = {}
        python_environment = PythonEnvironment.of("python")
        requirements = filter_dependencies(requirements, python_environment.pep508, extras)
        with tqdm.tqdm(desc="Resolving requirements graph") as progress:
            graph = build_distribution_graph(
                env=python_environment,
                dependencies=requirements,
                resolved_callback=lambda d: progress.update(len(d)),
                dists_cache=dists_cache,
            )

        graph.sort()
        output = t.cast(dict[str, t.Any], databind.json.dump(graph, DistributionGraph))

        # Retrieve the license text from the distributions.
        if self.option("with-license-text"):
            for dist_name, dist_data in output["metadata"].items():
                dist = dists_cache[dist_name]
                dist_data["license_text"] = None
                if dist is not None:
                    for filename in ("LICENSE", "LICENSE.txt", "LICENSE.text", "LICENSE.rst"):
                        try:
                            dist_data["license_text"] = dist.get_metadata(filename)
                            break
                        except FileNotFoundError:
                            pass

        print(json.dumps(output, indent=2, sort_keys=True))
        return 0


class ReportPlugin(ApplicationPlugin):
    def load_configuration(self, app: Application) -> None:
        return None

    def activate(self, app: Application, config: None) -> None:
        app.cleo.add(ReportDependenciesCommand(app))



import datetime
import textwrap
from pathlib import Path

from slap.application import Application, Command, argument, option
from slap.plugins import ApplicationPlugin
from slap.util.external.licenses import get_spdx_license_details, wrap_license_text
from slap.util.vcs import get_git_author

TEMPLATES = {
    "poetry": {
        "pyproject.toml": """
            [build-system]
            requires = ["poetry-core"]
            build-backend = "poetry.core.masonry.api"

            [tool.poetry]
            name = "{name}"
            version = "0.1.0"
            description = ""
            authors = ["{author_name} <{author_email}>"]
            license = "{license}"
            readme = "readme.md"
            packages = [{{ include = "{path}", from = "src" }}]
            classifiers = []
            keywords = []

            [tool.poetry.urls]
            # "Bug Tracker" = ""
            # Documentation = ""
            # Homepage = ""
            # Repository = ""

            [tool.poetry.dependencies]
            python = "^3.6"

            [tool.poetry.dev-dependencies]
            black = "*"
            flake8 = "*"
            isort = "*"
            mypy = "*"
            pytest = "*"

            [tool.slap]
            typed = true

            [tool.slap.test]
            check = "slap check"
            mypy = "dmypy run src/"
            pytest = "pytest tests/ -vv"
            black = "black --check src/ tests/"
            isort = "isort --check-only src/ tests/"
            flake8 = "flake8 src/ tests/"

            [tool.slap.run]
            fmt = "black src/ tests/ && isort src/ tests/"

            [tool.mypy]
            python_version = "3.6"
            explicit_package_bases = true
            mypy_path = ["src"]
            namespace_packages = true
            pretty = true
            show_error_codes = true
            show_error_context = true
            strict = true
            warn_no_return = true
            warn_redundant_casts = true
            warn_unreachable = true
            warn_unused_ignores = true

            [tool.isort]
            profile = "black"
            line_length = 120
            combine_as_imports = true

            [tool.black]
            line-length = 120
        """,
        "LICENSE": "",
        "readme.md": """
            # {name}

        """,
        ".flake8": """
            [flake8]
            max-line-length = 120
            # Black can yield formatted code that triggers these Flake8 warnings.
            ignore=
                W503, # line break before binary operator
                W504, # line break after binary operator
        """,
        ".gitignore": """
            /.vscode
            /dist
            /build
            .venv/
            *.egg-info/
            __pycache__/
            poetry.lock
        """,
        "src/{path}/__init__.py": """
            __version__ = "0.1.0"
        """,
        "tests/test_import.py": """
            def test_import():
                exec("from {dotted_name} import *")
    """,
        "src/{path}/py.typed": "",
    },
    "github": {
        ".github/workflows/python-package.yml": """
            name: Python application

            on:
              push: {{ branches: [ develop ], tags: [ "*" ] }}
              pull_request: {{ branches: [ develop ] }}

            jobs:
              test:
                runs-on: ubuntu-latest
                strategy:
                  fail-fast: false
                  matrix:
                    python-version: [ "3.6", "3.7", "3.8", "3.9", "3.10", "3.x" ]
                steps:
                - uses: actions/checkout@v2
                - uses: python-slap/slap-cli@gha/install/v1
                - uses: actions/setup-python@v2
                  with: {{ python-version: "${{{{ matrix.python-version }}}}" }}
                - run: slap install -vv --no-venv-check
                - run: slap test

              update-changelog:
                runs-on: ubuntu-latest
                if: github.event_name == 'pull_request'
                steps:
                  - uses: actions/checkout@v2
                  - uses: python-slap/slap-cli@gha/changelog/update/v1
    """,
    },
}


class InitCommandPlugin(ApplicationPlugin, Command):
    """Bootstrap some files for a Python project.

    Currently available templates:

    1. <info>poetry</info>
    """

    app: Application

    name = "init"
    arguments = [
        argument(
            "directory",
            description="The directory in which to create the generated files. If not specified, a new directory with "
            "the name specified via the <opt>--name</opt> option is created.",
            optional=True,
        )
    ]
    options = [
        option(
            "--name",
            description="The name of the Python package.",
            flag=False,
        ),
        option(
            "--license",
            description="The package license.",
            flag=False,
            default="MIT",
        ),
        option(
            "--template",
            "-t",
            description="The template to use.",
            flag=False,
            default="poetry",
        ),
        option(
            "--overwrite",
            "-f",
            description="Overwrite files.",
        ),
        option(
            "--dry",
            "-d",
            description="Dont actually write files.",
        ),
        option(
            "--as-markdown",
            description="Render the content as Markdown (uses by the Slap docs)",
        ),
    ]

    def load_configuration(self, app: Application) -> None:
        return None

    def activate(self, app: Application, config: None) -> None:
        self.app = app
        app.cleo.add(self)

    def handle(self) -> int:
        from nr.util.optional import Optional

        template = self.option("template")
        if template not in TEMPLATES:
            self.line_error(f'error: template "{template}" does not exist', "error")
            return 1

        author = get_git_author()
        name = self.option("name")
        directory = (
            Optional(self.argument("directory"))
            .map(Path)
            .map(lambda v: v or (name.replace(".", "-") if name else None))
            .or_else_get(Path.cwd)
        )

        scope = {
            "license": self.option("license"),
            "year": datetime.date.today().year,
            "author_name": author.name if author and author.name else "Unknown",
            "author_email": author.email if author and author.email else "me@unknown.org",
        }
        if self.option("name"):
            scope.update(
                {
                    "name": self.option("name"),
                    "dotted_name": self.option("name").replace("-", "_"),
                    "path": self.option("name").replace(".", "/").replace("-", "_"),
                    "package": self.option("name").replace(".", "_").replace("-", "_"),
                }
            )
        for filename, content in TEMPLATES[template].items():
            if filename == "LICENSE":
                content = get_spdx_license_details(self.option("license")).license_text
                content = wrap_license_text(content).replace("<year>", str(scope["year"]))
                content = wrap_license_text(content).replace("<copyright holders>", scope["author_name"])
            else:
                filename = filename.format(**scope)
                content = textwrap.dedent(content.format(**scope)).strip()
                if content:
                    content += "\n"

            path = directory / filename

            if self.option("as-markdown"):
                print(f'```{path.suffix[1:]} title="{path}"')
                print(content, "    ")
                print("```\n\n")
                continue

            if path.exists() and not self.option("overwrite"):
                self.line(f"skip <info>{path}</info> (already exists)")
                continue

            if not self.option("dry") and not self.option("as-markdown"):
                path.parent.mkdir(parents=True, exist_ok=True)
                path.write_text(content)

            self.line(f"write <info>{path}</info>")

        return 0



from __future__ import annotations

import dataclasses
import functools
import json
import logging
import pickle
import subprocess as sp
import textwrap
import typing as t
from pathlib import Path

from slap.python import pep508

if t.TYPE_CHECKING:
    import pkg_resources

    from slap.python.dependency import Dependency

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class PythonEnvironment:
    """Represents a Python environment. Provides functionality to introspect the environment."""

    executable: str
    version: str
    version_tuple: tuple[int, int, int]
    platform: str
    prefix: str
    base_prefix: str | None
    real_prefix: str | None
    pep508: pep508.Pep508Environment
    _has_pkg_resources: bool | None = None

    def is_venv(self) -> bool:
        """Checks if the Python environment is a virtual environment."""

        return bool(self.real_prefix or (self.base_prefix and self.prefix != self.base_prefix))

    def has_pkg_resources(self) -> bool:
        """Checks if the Python environment has the `pkg_resources` module available."""

        if self._has_pkg_resources is None:
            code = textwrap.dedent(
                """
        try: import pkg_resources
        except ImportError: print('false')
        else: print('true')
      """
            )
            self._has_pkg_resources = json.loads(sp.check_output([self.executable, "-c", code]).decode())
        return self._has_pkg_resources

    @staticmethod
    @functools.lru_cache()
    def of(python: str | t.Sequence[str]) -> "PythonEnvironment":
        """Introspects the given Python installation to construct a #PythonEnvironment."""

        if isinstance(python, str):
            python = [python]

        # We ensure that the Pep508 module is importable.
        pep508_path = str(Path(pep508.__file__).parent)

        code = textwrap.dedent(
            f"""
      import sys, platform, json, pickle
      sys.path.append({pep508_path!r})
      import pep508
      try: import pkg_resources
      except ImportError: pkg_resources = None
      print(json.dumps({{
        "executable": sys.executable,
        "version": sys.version,
        "version_tuple": sys.version_info[:3],
        "platform": platform.platform(),
        "prefix": sys.prefix,
        "base_prefix": getattr(sys, 'base_prefix', None),
        "real_prefix": getattr(sys, 'real_prefix', None),
        "pep508": pep508.Pep508Environment.current().as_json(),
        "_has_pkg_resources": pkg_resources is not None,
      }}))
    """
        )

        payload = json.loads(sp.check_output(list(python) + ["-c", code]).decode())
        payload["version_tuple"] = tuple(payload["version_tuple"])
        payload["pep508"] = pep508.Pep508Environment(**payload["pep508"])
        return PythonEnvironment(**payload)

    def get_distribution(self, distribution: str) -> pkg_resources.Distribution | None:
        """Query the details for a single distribution in the Python environment."""

        return self.get_distributions([distribution])[distribution]

    def get_distributions(self, distributions: t.Collection[str]) -> dict[str, pkg_resources.Distribution | None]:
        """Query the details for the given distributions in the Python environment with
        #pkg_resources.get_distribution()."""

        code = textwrap.dedent(
            """
      import sys, pkg_resources, pickle
      result = []
      for arg in sys.argv[1:]:
        try:
          dist = pkg_resources.get_distribution(arg)
        except pkg_resources.DistributionNotFound:
          dist = None
        result.append(dist  )
      sys.stdout.buffer.write(pickle.dumps(result))
    """
        )

        keys = list(distributions)
        result = pickle.loads(sp.check_output([self.executable, "-c", code] + keys))
        return dict(zip(keys, result))


@dataclasses.dataclass
class DistributionMetadata:
    """Additional metadata for a distribution."""

    location: str
    version: str
    license_name: str | None
    platform: str | None
    requires_python: str | None
    requirements: list[str]
    extras: set[str]


def get_distribution_metadata(dist: pkg_resources.Distribution) -> DistributionMetadata:
    """Parses the distribution metadata."""

    from email.parser import Parser

    data = Parser().parsestr(dist.get_metadata(dist.PKG_INFO))

    return DistributionMetadata(
        location=dist.location,
        version=data.get("Version"),
        license_name=data.get("License"),
        platform=data.get("Platform"),
        requires_python=data.get("Requires-Python"),
        requirements=data.get_all("Requires-Dist") or [],
        extras=set(data.get_all("Provides-Extra") or []),
    )


@dataclasses.dataclass
class DistributionGraph:
    """Represents a resolved graph of distributions, their metadata and dependencies in a Python environment."""

    #: Maps a distribution name to it's metadata.
    metadata: dict[str, DistributionMetadata]

    #: Maps out the dependencies between distributions.
    dependencies: dict[str, t.MutableSet[str]]

    #: A set of the distributions that have been found to be required but could not be resolved.
    missing: set[str]

    def sort(self) -> None:
        from nr.util.orderedset import OrderedSet

        for dist_name, dependencies in self.dependencies.items():
            self.dependencies[dist_name] = OrderedSet(sorted(dependencies))

    def update(self, other: DistributionGraph) -> None:
        self.metadata.update(other.metadata)
        self.dependencies.update(other.dependencies)
        self.missing.update(other.missing)


def build_distribution_graph(
    env: PythonEnvironment,
    dependencies: list[Dependency],
    resolved_callback: t.Callable[[dict[str, pkg_resources.Distribution | None]], t.Any] | None = None,
    dists_cache: dict[str, pkg_resources.Distribution | None] | None = None,
) -> DistributionGraph:
    """Builds a #DistributionGraph in the given #PythonEnvironment using the given dependencies.

    Args:
      env: The Python environment in which to resolve the dependencies.
      dependencies: The dependencies to resolve. Note that this list should already be filtered by its markers.
      resolved_callback: A callback that is invoked with the list of dependencies that have been successfully
        resolved. This is useful for progress reporting.
    """

    from slap.python.dependency import parse_dependencies
    from slap.python.pep508 import filter_dependencies

    graph = DistributionGraph({}, {}, set())

    if dists_cache is None:
        dists_cache = {}

    logger.info("Fetching requirements: <val>%s</val>", dependencies)

    # TODO (@NiklasRosenstein): Should we warn here if the dependencies map is smaller than the dependencies list?
    dependencies_map = {dependency.name: dependency for dependency in dependencies}

    # Resolve the distributions available in the Python environment.
    distributions = {dist_name: dists_cache[dist_name] for dist_name in dependencies_map if dist_name in dists_cache}
    fetch_distributions = dependencies_map.keys() - distributions.keys()
    if fetch_distributions:
        fetched_distributions = env.get_distributions(fetch_distributions)
        distributions.update(fetched_distributions)
        dists_cache.update(fetched_distributions)

    if resolved_callback:
        resolved_callback(distributions)

    # Parse the dependencies of the distributions.
    prefetch_distributions: set[str] = set()
    metadata: dict[str, tuple[DistributionMetadata, list[Dependency]]] = {}
    for dist_name, dist in distributions.items():
        if dist is None:
            graph.missing.add(dist_name)
        else:
            dist_meta = get_distribution_metadata(dist)
            dist_extras = set(dependencies_map[dist_name].extras or [])
            parsed_dependencies = filter_dependencies(
                parse_dependencies(dist_meta.requirements), env.pep508, dist_extras
            )
            metadata[dist_name] = (dist_meta, parsed_dependencies)
            prefetch_distributions |= {dependency.name for dependency in parsed_dependencies}

    # Prefetch distributions.
    prefetch_distributions -= dists_cache.keys()
    if prefetch_distributions:
        dists_cache.update(env.get_distributions(prefetch_distributions))

    # Continue building the graph recursively.
    for dist_name, (dist_meta, parsed_dependencies) in metadata.items():
        graph.metadata[dist_name] = dist_meta

        for dependency in parsed_dependencies:
            graph.dependencies.setdefault(dist_name, set()).add(dependency.name)

        # TODO (@NiklasRosenstein): Potential optimization here is to cache which distributions have already been
        #   resolved; and to pre-fetch all distributions needed in the recursive call in the parent collectively
        #   across all the dependencies of the current set of distributions.
        new_graph = build_distribution_graph(
            env=env,
            dependencies=parsed_dependencies,
            resolved_callback=resolved_callback,
            dists_cache=dists_cache,
        )

        graph.update(new_graph)

    return graph



from __future__ import annotations

import os
import shlex
import typing as t
from pathlib import Path

from slap.application import Application, Command
from slap.plugins import ApplicationPlugin

if t.TYPE_CHECKING:
    from slap.python.dependency import Dependency


class InfoCommandPlugin(Command, ApplicationPlugin):
    """Show info about the Slap application workspace and the loaded projects."""

    app: Application
    name = "info"

    def load_configuration(self, app: Application) -> None:
        return None

    def activate(self, app: Application, config: None) -> None:
        self.app = app
        app.cleo.add(self)

    def handle(self) -> int:
        projects = self.app.repository.get_projects_ordered()

        self.line(f'Repository <s>"{self.app.repository.directory}"</s>')
        self.line(f"  vcs: <opt>{self.app.repository.vcs()}</opt>")
        self.line(f"  host: <opt>{self.app.repository.host()}</opt>")
        self.line(f"  projects: <opt>{[p.id for p in projects]}</opt>")

        for project in projects:
            if not project.is_python_project:
                continue
            packages_list = project.packages()
            packages = (
                "<i>none</i>"
                if packages_list is None
                else "[]"
                if len(packages_list or []) == 0
                else ", ".join(
                    f"<opt>{p.name} ({os.path.relpath(p.root, project.directory)})</opt>" for p in packages_list
                )
            )
            self.line(
                f'Project <s>"{os.path.relpath(project.directory, Path.cwd())}" (id: <opt>{project.id}</opt>)</s>'
            )
            self.line(f"  version: <opt>{project.version()}</opt>")
            self.line(f"  dist-name: <opt>{project.dist_name()}</opt>")
            self.line(f"  packages: {packages}")
            self.line(f"  readme: <opt>{project.handler().get_readme(project)}</opt>")
            self.line(f"  handler: <opt>{project.handler()}</opt>")

            inter_deps = project.get_interdependencies(projects)
            if inter_deps:
                project_names = ", ".join(f"<opt>{p.dist_name()}</opt>" for p in inter_deps)
                self.line(f"  depends on: {project_names}")

            deps = project.dependencies()
            self.line("  dependencies:")
            self._print_deps("run", deps.run)
            self._print_deps("dev", deps.dev)
            for key, value in deps.extra.items():
                self._print_deps(f"extra.{key}", value)

        return 0

    def _print_deps(self, prefix: str, deps: t.Sequence[Dependency]) -> None:
        from slap.install.installer import PipInstaller

        if deps:
            self.line(f"    {prefix}:")
            for dep in sorted(deps, key=lambda s: s.name.lower()):
                self.line(
                    f'      - <opt>{" ".join(map(shlex.quote, PipInstaller.dependency_to_pip_arguments(dep)))}</opt>'
                )
        else:
            self.line(f"    {prefix}: <i>none</i>")



""" Project handler for projects using the Flit build system. """

from __future__ import annotations

import logging
import typing as t

from slap.ext.project_handlers.base import PyprojectHandler

if t.TYPE_CHECKING:
    from slap.project import Dependencies, Project
    from slap.python.dependency import Dependency

logger = logging.getLogger(__name__)


class FlitProjectHandler(PyprojectHandler):

    # ProjectHandlerPlugin

    def matches_project(self, project: Project) -> bool:
        if not project.pyproject_toml.exists():
            return False
        build_backend = project.pyproject_toml.get("build-system", {}).get("build-backend")
        return build_backend == "flit_core.buildapi"

    def get_dist_name(self, project: Project) -> str | None:
        return project.pyproject_toml.get("tool", {}).get("flit", {}).get("metadata", {}).get("module", {}).get("name")

    def get_readme(self, project: Project) -> str | None:
        return (
            project.pyproject_toml.get("project", {}).get("readme")
            or project.pyproject_toml.get("tool", {}).get("flit", {}).get("metadata", {}).get("description-file")
            or super().get_readme(project)
        )

    def get_dependencies(self, project: Project) -> Dependencies:
        from slap.project import Dependencies
        from slap.python.dependency import PypiDependency, VersionSpec

        flit: dict[str, t.Any] | None = project.pyproject_toml.get("tool", {}).get("flit")
        project_conf: dict[str, t.Any] | None = project.pyproject_toml.get("project")
        build_dependencies = PypiDependency.parse_list(
            project.pyproject_toml.get("build-system", {}).get("requires", [])
        )

        if project_conf is not None:
            optional = project_conf.get("optional-dependencies", {})
            return Dependencies(
                VersionSpec(project_conf["requires-python"]) if "requires-python" in project_conf else None,
                PypiDependency.parse_list(project_conf.get("dependencies", [])),
                PypiDependency.parse_list(optional.pop("dev", [])),
                {extra: PypiDependency.parse_list(value) for extra, value in optional.items()},
                build_dependencies,
            )
        elif flit is not None:
            optional = flit.get("requires-extra", {})
            return Dependencies(
                VersionSpec(flit["requires-python"]) if "requires-python" in flit else None,
                PypiDependency.parse_list(flit.get("requires", [])),
                PypiDependency.parse_list(optional.pop("dev", [])),
                {extra: PypiDependency.parse_list(value) for extra, value in optional.items()},
                build_dependencies,
            )
        else:
            logger.warning("Unable to read dependencies for project <subj>%s</subj>", project)
            return Dependencies(None, [], [], {}, build_dependencies)

    def get_add_dependency_toml_location_and_config(
        self,
        project: Project,
        dependency: Dependency,
        where: str,
    ) -> tuple[list[str], list | dict]:
        from slap.python.dependency import PypiDependency

        if not isinstance(dependency, PypiDependency):
            raise Exception(f"Flit project handler only supports PypiDependency, got {dependency!r}")

        flit: dict[str, t.Any] | None = project.pyproject_toml.get("tool", {}).get("flit")
        if flit is not None:
            locator = ["requires"] if where == "run" else ["requires-extras", where]
            return ["tool", "flit"] + locator, [dependency.version.to_pep_508()]
        else:
            locator = ["dependencies"] if where == "run" else ["optional-dependencies", where]
            return ["project"] + locator, [dependency.version.to_pep_508()]



import contextlib
import tempfile
from pathlib import Path
from typing import Iterable

import build
import build.env

from slap.application import Application, Command, option
from slap.install.installer import PipInstaller
from slap.plugins import ApplicationPlugin


def flatten(it: Iterable[Iterable[str]]) -> Iterable[str]:
    for item in it:
        yield from item


class PublishCommandPlugin(Command, ApplicationPlugin):
    """A wrapper to publish the Python project to a repository such as PyPI.

    Uses the PEP 517 build system defined in the <code>pyproject.toml</code> to build
    packages and then uploads them with Twine. Note that it currently expects the build
    backend to be installed already.

    The command-line options are almost identical to the <code>twine upload</code> command.

    Note: You can combine the `-d` and `-b` options to effectively perform a build, storing
    the artifacts into the specified directory but not publishing them.
    """

    app: Application

    name = "publish"
    options = [
        option(
            "python",
            flag=False,
            description="use this Python executable to build the distribution but do not automatically install build "
            "requirements into it; if not specified a temporary build environment is created",
        ),
        option("repository", "r", flag=False, default="pypi"),
        option("repository-url", flag=False),
        option("sign", "s"),
        option("sign-with", flag=False),
        option("identity", "i", flag=False),
        option("username", "u", flag=False),
        option("password", "p", flag=False),
        option("non-interactive"),
        option("comment", "c", flag=False),
        option("config-file", flag=False, default="~/.pypirc"),
        option("skip-existing"),
        option("cert", flag=False),
        option("client-cert", flag=False),
        # option("verbose"),
        option("disable-progress-bar"),
        option("dry", "d"),
        option("build-directory", "b", flag=False),
    ]

    def load_configuration(self, app: Application) -> None:
        return None

    def activate(self, app: Application, config: None) -> None:
        self.app = app
        return app.cleo.add(self)

    def handle(self) -> int:
        from twine.commands.upload import upload
        from twine.settings import Settings

        distributions: list[Path] = []

        with contextlib.ExitStack() as stack:
            build_dir = self.option("build-directory")
            if build_dir is None:
                build_dir = stack.enter_context(tempfile.TemporaryDirectory())

            executable = self.option("python")
            if not executable:
                isolated_env = stack.enter_context(build.env.IsolatedEnvBuilder())
                executable = isolated_env.executable
            else:
                isolated_env = None

            for project in self.app.repository.projects():
                if isolated_env:
                    isolated_env.install(
                        list(flatten(PipInstaller.dependency_to_pip_arguments(x) for x in project.dependencies().build))
                    )
                if not project.is_python_project:
                    continue

                self.line(f"Build <info>{project.dist_name()}</info>")
                builder = build.ProjectBuilder(project.directory, executable)

                sdist = Path(builder.build("sdist", build_dir))
                self.line(f"  <comment>{sdist.name}</comment>")
                wheel = Path(builder.build("wheel", build_dir))
                self.line(f"  <comment>{wheel.name}</comment>")

                distributions += [sdist, wheel]

            if not self.option("dry"):
                self.line("Publishing")
                kwargs = {option.name.replace("-", "_"): self.option(option.name) for option in self.options}
                kwargs["repository_name"] = kwargs.pop("repository")
                settings = Settings(**kwargs)
                upload(settings, [str(d) for d in distributions])

        return 0



import logging

from slap.plugins import ReleasePlugin
from slap.project import Project
from slap.release import VersionRef, match_version_ref_pattern

logger = logging.getLogger(__name__)


class SourceCodeVersionReferencesPlugin(ReleasePlugin):
    """This plugin searches for a `__version__` key in the source code of the project and return it as a version
    reference. Based on the Poetry configuration (considering `tool.poetry.packages` and searching in the `src/`
    folder if it exists), the following source files will be checked:

    * `__init__.py`
    * `__about__.py`
    * `_version.py`

    Note that configuring `tool.poetry.packages` is needed for the detection to work correctly with PEP420
    namespace packages.
    """

    VERSION_REGEX = r'^__version__\s*=\s*[\'"]([^\'"]+)[\'"]'
    FILENAMES = ["__init__.py", "__about__.py", "_version.py", ".py"]

    def get_version_refs(self, project: Project) -> list[VersionRef]:
        packages = project.packages()
        if not packages:
            return []

        results = []
        packages_without_version = []
        for package in packages:
            for path in [package.path] if package.path.is_file() else [package.path / f for f in self.FILENAMES]:
                if path.exists():
                    try:
                        version_ref = match_version_ref_pattern(path, self.VERSION_REGEX)
                    except ValueError as exc:
                        logger.warning("%s", exc)
                        continue
                    if version_ref:
                        results.append(version_ref)
                        break
            else:
                packages_without_version.append(package)

        if packages_without_version:
            logger.warning(
                "Unable to detect <val>__version__</val> in the following packages of project "
                "<subj>%s</sub>: <obj>%s</obj>",
                project,
                [p.name for p in packages_without_version],
            )

        return results



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['slap',
 'slap.ext',
 'slap.ext.application',
 'slap.ext.changelog_update_automation',
 'slap.ext.checks',
 'slap.ext.project_handlers',
 'slap.ext.release',
 'slap.ext.repository_handlers',
 'slap.ext.repository_hosts',
 'slap.install',
 'slap.python',
 'slap.util',
 'slap.util.external']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=4.0',
 'beautifulsoup4>=4.10.0,<5.0.0',
 'build>=0.8.0,<0.9.0',
 'cleo>=1.0.0a4',
 'databind>=2.0.0,<3.0.0',
 'flit>=3.6.0,<4.0.0',
 'nr.python.environment>=0.1.4,<0.2.0',
 'nr.util>=0.8.12,<1.0.0',
 'poetry-core>=1.1.0a6,<2.0.0',
 'ptyprocess>=0.7.0,<0.8.0',
 'pygments>=2.11.2,<3.0.0',
 'requests>=2.27.1,<3.0.0',
 'setuptools>=39.1.0',
 'tomli>=2.0.0,<3.0.0',
 'tomlkit>=0.10.1,<0.11.0',
 'tqdm>=4.64.0,<5.0.0',
 'twine>=3.7.0,<4.0.0']

entry_points = \
{'console_scripts': ['slap = slap.__main__:main'],
 'slap.plugins.application': ['add = slap.ext.application.add:AddCommandPlugin',
                              'changelog = '
                              'slap.ext.application.changelog:ChangelogCommandPlugin',
                              'check = '
                              'slap.ext.application.check:CheckCommandPlugin',
                              'info = '
                              'slap.ext.application.info:InfoCommandPlugin',
                              'init = '
                              'slap.ext.application.init:InitCommandPlugin',
                              'install = '
                              'slap.ext.application.install:InstallCommandPlugin',
                              'link = '
                              'slap.ext.application.link:LinkCommandPlugin',
                              'publish = '
                              'slap.ext.application.publish:PublishCommandPlugin',
                              'release = '
                              'slap.ext.application.release:ReleaseCommandPlugin',
                              'report = '
                              'slap.ext.application.report:ReportPlugin',
                              'run = slap.ext.application.run:RunCommandPlugin',
                              'test = '
                              'slap.ext.application.test:TestCommandPlugin',
                              'venv = slap.ext.application.venv:VenvPlugin'],
 'slap.plugins.changelog_update_automation': ['github-actions = '
                                              'slap.ext.changelog_update_automation.github_actions:GithubActionsChangelogUpdateAutomationPlugin'],
 'slap.plugins.check': ['changelog = '
                        'slap.ext.checks.changelog:ChangelogValidationCheckPlugin',
                        'general = slap.ext.checks.general:GeneralChecksPlugin',
                        'poetry = slap.ext.checks.poetry:PoetryChecksPlugin',
                        'release = '
                        'slap.ext.checks.release:ReleaseChecksPlugin'],
 'slap.plugins.project': ['flit = '
                          'slap.ext.project_handlers.flit:FlitProjectHandler',
                          'poetry = '
                          'slap.ext.project_handlers.poetry:PoetryProjectHandler',
                          'setuptools = '
                          'slap.ext.project_handlers.setuptools:SetuptoolsProjectHandler'],
 'slap.plugins.release': ['changelog_release = '
                          'slap.ext.release.changelog:ChangelogReleasePlugin',
                          'source_code_version = '
                          'slap.ext.release.source_code_version:SourceCodeVersionReferencesPlugin'],
 'slap.plugins.repository': ['default = '
                             'slap.ext.repository_handlers.default:DefaultRepositoryHandler'],
 'slap.plugins.repository_host': ['github = '
                                  'slap.ext.repository_hosts.github:GithubRepositoryHost'],
 'slap.plugins.version_incrementing_rule': ['major = '
                                            'slap.ext.version_incrementing_rule:major',
                                            'minor = '
                                            'slap.ext.version_incrementing_rule:minor',
                                            'patch = '
                                            'slap.ext.version_incrementing_rule:patch',
                                            'premajor = '
                                            'slap.ext.version_incrementing_rule:premajor',
                                            'preminor = '
                                            'slap.ext.version_incrementing_rule:preminor',
                                            'prepatch = '
                                            'slap.ext.version_incrementing_rule:prepatch',
                                            'prerelease = '
                                            'slap.ext.version_incrementing_rule:prerelease']}

setup_kwargs = {
    'name': 'slap-cli',
    'version': '1.6.30',
    'description': 'Slap is a command-line utility for developing Python applications.',
    'long_description': '# Slap\n\n<img src="docs/content/img/logo.svg" style="height: 200px !important">\n\n  [PEP 517]: https://peps.python.org/pep-0517/\n\nSlap is a command-line tool to simplify command workflows in the development of Python projects\nindependent of the [PEP 517][] build backend being used, capable of managing single- and multi-project\nrepositories.\n\n## Installation\n\nI recommend installing Slap using Pipx. (Requires Python 3.10 or higher)\n\n    $ pipx install slap-cli\n\n> __Note__: Currently Slap relies on an alpha version of `poetry-core` (`^1.1.0a6`). If you install it into\n> the same environment as Poetry itself, you may also need to use an alpha version of Poetry (e.g. `1.2.0a2`).\n>\n> If you use Slap in GitHub Actions, try one of the actions provided by Slap directly:\n>\n> * [`NiklasRosenstein/slap@gha/install/v1`](https://python-slap.github.io/slap-cli/guides/github/#install-slap)\n> * [`NiklasRosenstein/slap@gha/changelog/update/v1`](https://python-slap.github.io/slap-cli/guides/github/#update-changelogs)\n\n## Documentation\n\nYou can find the documentation for Slap here: <https://python-slap.github.io/slap-cli/>\n\nCheck out the [Getting started](https://python-slap.github.io/slap-cli/getting-started/) guide.\n\n## Feature Matrix\n\n| Feature | Poetry | Documentation |\n| ------- | ------ | ------------- |\n| Manage structured changelog entries | ❌ | [slap changelog](https://python-slap.github.io/slap-cli/commands/changelog/) |\n| Show project details | ❌ | [slap info](https://python-slap.github.io/slap-cli/commands/info/) |\n| Build and publish to PyPI using Twine | ✅ (single project only) | [slap publish](https://python-slap.github.io/slap-cli/commands/publish/) |\n| Create a new release (bump version numbersr)| ❌ (sub-par support) | [slap release](https://python-slap.github.io/slap-cli/commands/release/) |\n| Run a command configured in `pyproject.toml` | ❌ | [slap run](https://python-slap.github.io/slap-cli/commands/run/) |\n| Run tests configured in `pyproject.toml` | ❌ | [slap test](https://python-slap.github.io/slap-cli/commands/test/) |\n| Manage Python virtualenv\'s | ✅ (but out-of-worktree) | [slap venv](https://python-slap.github.io/slap-cli/commands/venv/) |\n| Generate a dependencies report | ❌ | [slap report dependencies](https://python-slap.github.io/slap-cli/commands/report/) |\n\n| Feature / Build backend | Flit  | Poetry  | Setuptools  | Documentation |\n| ----------------------- | ----- | ------- | ----------- | --------- |\n| Add dependency | ✅ | ✅ | ❌ | [slap add](https://python-slap.github.io/slap-cli/commands/add/) |\n| Sanity check project configuration | | ✅ | | [slap check](https://python-slap.github.io/slap-cli/commands/check/) |\n| Bootstrap project files | | ✅ | | [slap init](https://python-slap.github.io/slap-cli/commands/init/) |\n| Install projects using Pip | ✅ | ✅ | ✅ | [slap install](https://python-slap.github.io/slap-cli/commands/install/) |\n| Symlink projects (editable installs) | ✅ | ✅ | ✅ | [slap link](https://python-slap.github.io/slap-cli/commands/link/) |\n| Bump interdependencies in mono-repository | ✅ (not tested regularly) | ✅ | ✅ (partial) | [slap release](https://python-slap.github.io/slap-cli/commands/release/) |\n\n> __Legend__: ✅ explicitly supported, ❌ explicitly not supported, (blank) not relevant or currently not supported\n\n## Issues / Suggestions / Contributions\n\n  [GitHub Issues]: https://github.com/NiklasRosenstein/slap/issues\n  [GitHub Discussions]: https://github.com/NiklasRosenstein/slap/discussions\n  [GitHub Repository]: https://github.com/NiklasRosenstein/slap\n\nSlap is currently very opinionated by the fact that I built it as my personal workflow tool, but I welcome\nsuggestions and contributions, and I am hopeful it will be useful to a wider audience than myself.\n\nPlease report any issues you encounter via [GitHub Issues][]. Feel free to use the [GitHub Discussions][] forum\nto ask questions or make suggestions on new features (e.g. if you would like a new build backend to be supported?).\nLastly, feel free to submit pull requests to the [GitHub Repository][].\n\n## FAQ\n\n### Why "Slap"?\n\nFinding a good, catchy name that also types easily in the terminal and is not already widely used isn\'t easy, ok?\n\n### What makes this different to the Poetry CLI?\n\nSome people might find this similar to tools like Poetry, and while there is some overlap in functionality, Slap is\n**not a build backend** and is more targeted towards library development. In fact, most of my projects use Poetry as\nthe build backend but I never even once interact with the Poetry CLI throughout the lifetime of the project.\n\nThe most notable differences to Poetry are\n\n* Supports mono-repositories (i.e. multiple related Python projects in the same repository), to the extent that it\n  bumps version numbers of project inter-dependencies and installs your projects in topological order\n* Supports development installs independent of the build backend (yes; this means you can install Poetry packages\n  in editable mode even though the Poetry backend right now does not support editable installs)\n* Slap\'s version bump command (`slap release`) updates the version not just in your `pyproject.toml` but also the\n  `__version__` in your source code as well as in related projects (see mono-repositories above) and any additional\n  references you can configure via Regex patterns\n* Does not automagically create a virtual environment for you when instal your project(s); instead, it errors when\n  you try to install into a non-virtual Python environment and gives you an easy-to-use tool to create and activate\n  virtual environments (and allowing multiple environments per project as well as global environments)\n* Uses Pip to install your project(s), unlike Poetry which comes with its own dependency resolver and package\n  installer (which I personally have been having a lot of issues with in the past).\n* Does not have a concept of lock files\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import abc
import copy
import dataclasses
import datetime
import typing as t
import uuid
from pathlib import Path

from databind.core.settings import Alias
from nr.util.weak import weak_property

if t.TYPE_CHECKING:
    from poetry.core.semver.version import Version  # type: ignore[import]

    from slap.repository import RepositoryHost


def is_url(s: str) -> bool:
    return s.startswith("http://") or s.startswith("https://")


@dataclasses.dataclass
class ChangelogEntry:
    id: str
    type: str
    description: str
    author: str | None = None
    authors: list[str] | None = None
    pr: str | None = None
    issues: list[str] | None = None

    def get_authors(self) -> list[str]:
        result = []
        if self.author is not None:
            result.append(self.author)
        if self.authors is not None:
            result += self.authors
        return result


@dataclasses.dataclass
class Changelog:
    entries: list[ChangelogEntry] = dataclasses.field(default_factory=list)
    release_date: t.Annotated[datetime.date | None, Alias("release-date")] = None


class ChangelogDeser(abc.ABC):
    @abc.abstractmethod
    def load(self, fp: t.TextIO, filename: str) -> Changelog:
        ...

    def save(self, changelog: Changelog, fp: t.TextIO, filename: str) -> None:
        fp.write(self.dump(changelog))

    @abc.abstractmethod
    def dump(self, changelog: Changelog) -> str:
        ...

    @abc.abstractmethod
    def dump_entry(self, entry: ChangelogEntry) -> str:
        ...


class TomlChangelogDeser(ChangelogDeser):
    def load(self, fp: t.TextIO, filename: str) -> Changelog:
        import databind.json
        import tomli

        return databind.json.load(tomli.loads(fp.read()), Changelog, filename=filename)

    def dump(self, changelog: Changelog) -> str:
        import databind.json
        import tomli_w
        from databind.core.settings import SerializeDefaults

        data = databind.json.dump(changelog, Changelog, settings=[SerializeDefaults(False)])
        return tomli_w.dumps(t.cast(dict, data))

    def dump_entry(self, entry: ChangelogEntry) -> str:
        import databind.json
        import tomli_w
        from databind.core.settings import SerializeDefaults

        return tomli_w.dumps(
            t.cast(dict, databind.json.dump(entry, ChangelogEntry, settings=[SerializeDefaults(False)]))
        )


class ManagedChangelog:

    _manager: "ChangelogManager" = weak_property("_ManagedChangelog__manager")

    def __init__(self, manager: "ChangelogManager", path: Path, version: str | None) -> None:
        from poetry.core.semver.version import Version

        assert version is None or isinstance(version, str), type(version)

        self.path = path
        self.version = Version.parse(version) if version else None
        self._manager = manager
        self._content: Changelog | None = None

    @property
    def content(self) -> Changelog:
        return self.load()

    def exists(self) -> bool:
        return self.path.exists()

    def load(self, reload: bool = False) -> Changelog:
        if self._content is None or reload:
            self._content = self._manager._load(self.path)
        return self._content

    def save(self, changelog: Changelog | None) -> None:
        if changelog is None:
            if self._content is None:
                raise RuntimeError('ManagedChangelog.content was not loaded and no "changelog" parameter was provided')
            changelog = self._content
        if changelog.release_date is None and self.path.name != self._manager.unreleased_fn:
            raise RuntimeError(
                f"changelog without release date must be the unreleased changelog (but is {self.path.name})"
            )
        if changelog.release_date is not None and self.path.name == self._manager.unreleased_fn:
            raise RuntimeError(f"changelog with release date must be a version (but is {self.path.name})")
        self._manager._save(changelog, self.path)

    def release(self, version: str) -> None:
        """Releases the changelog as the specified version."""

        if self.version:
            raise RuntimeError("cannot release already released changelog")

        content = copy.deepcopy(self.content)
        content.release_date = datetime.date.today()
        target = self._manager.version(version)
        target.save(content)
        self.path.unlink()


@dataclasses.dataclass
class ChangelogManager:
    """Manages a directory of changelogs."""

    #: The directory in which the changelog files can be found.
    directory: Path

    #: An instance for validation and normalization of issue and PR references.
    repository_host: RepositoryHost | None

    #: The name of the file that contains the unreleased changes.
    unreleased_fn: str = "_unreleased.toml"

    #: The template to describe the filenames of released changedlogs.
    version_fn_template: str = "{version}.toml"

    #: A list of strings that represent the valid choices of changelog entry types.
    valid_types: list[str] | None = None

    #: The de/serializer for changelogs.
    deser: ChangelogDeser = dataclasses.field(default_factory=TomlChangelogDeser)

    #: If enabled, write operations to the changelog directory are disabled.
    readonly: bool = False

    def _load(self, file: Path) -> Changelog:
        with file.open("r") as fp:
            return self.deser.load(fp, str(file))

    def _save(self, changelog: Changelog, file: Path) -> None:
        if self.readonly:
            raise RuntimeError(f'"{self.directory}" is readonly')
        file.parent.mkdir(parents=True, exist_ok=True)
        with file.open("w") as fp:
            self.deser.save(changelog, fp, str(file))

    def unreleased(self) -> ManagedChangelog:
        return ManagedChangelog(self, self.directory / self.unreleased_fn, None)

    def version(self, version: str) -> ManagedChangelog:
        return ManagedChangelog(self, self.directory / self.version_fn_template.format(version=version), version)

    def all(self) -> list[ManagedChangelog]:
        if not self.directory.exists():
            return []
        changelogs = []
        for path in self.directory.iterdir():
            if path.suffix == ".toml" and path.name != self.unreleased_fn:
                changelogs.append(ManagedChangelog(self, path, path.stem))
        changelogs.sort(key=lambda c: t.cast("Version", c.version), reverse=True)
        unreleased = self.unreleased()
        if unreleased.exists():
            changelogs.insert(0, unreleased)
        return changelogs

    def make_entry(
        self,
        change_type: str,
        description: str,
        author: str,
        pr: str | None,
        issues: list[str] | None,
    ) -> ChangelogEntry:
        """Creates a new #ChangelogEntry and validates it. If the parameters of the changelog are invalid, a
        #InvalidChangelogEntryException is raised. A random unique ID is generated for the changelog. The *pr*
        and *issues* parameters may be issue IDs if a #remote is configured that supports the conversion. If no
        author is specified, it will be read from the *author* option or otherwise obtained via the #RepositoryHost,
        if available."""

        if self.valid_types is not None and change_type not in self.valid_types:
            raise ValueError(f"invalid change type: {change_type}")

        if pr is not None and self.repository_host:
            pr = self.repository_host.get_pull_request_by_reference(pr).url
        if issues is not None and self.repository_host:
            issues = [self.repository_host.get_issue_by_reference(i).url or i for i in issues]

        changelog_id = str(uuid.uuid4())
        return ChangelogEntry(
            id=changelog_id, type=change_type, description=description, author=author, pr=pr, issues=issues or None
        )

    def validate_entry(self, entry: ChangelogEntry) -> None:
        if self.valid_types is not None and entry.type not in self.valid_types:
            raise ValueError(f"invalid change type: {entry.type}")
        if entry.authors is not None and entry.author is not None:
            raise ValueError('entry has "author" and "authors", only one should be present')
        if not entry.get_authors():
            raise ValueError('entry has no "author" or "authors"')
        if not all(entry.get_authors()):
            raise ValueError("empty string in author(s)")
        if self.repository_host:
            if entry.pr:
                entry.pr = self.repository_host.get_pull_request_by_reference(entry.pr).url
            if entry.issues:
                entry.issues = [self.repository_host.get_issue_by_reference(issue).url for issue in entry.issues]



import logging

from slap.application import Application, argument, option
from slap.ext.application.install import get_active_python_bin, python_option, venv_check
from slap.ext.application.venv import VenvAwareCommand
from slap.plugins import ApplicationPlugin
from slap.python.dependency import PypiDependency, VersionSpec

logger = logging.getLogger(__name__)


class AddCommandPlugin(VenvAwareCommand, ApplicationPlugin):
    """Add one or more dependencies to a project."""

    app: Application
    name = "add"

    arguments = [
        argument(
            "packages",
            description="One or more packages to install with Pip and add to the project configuration.",
            multiple=True,
        )
    ]
    options = VenvAwareCommand.options + [
        option(
            "--dev",
            "-d",
            description="Add as development dependencies.",
        ),
        option(
            "--extra",
            "-e",
            description="Add as extra dependencies for the specified extra name.",
            flag=False,
        ),
        option(
            "--no-install",
            description="Do not actually install the dependencies with Pip. Note that if the dependency is not already "
            "installed and no version selector is specified with the package name, it will fall back to a match-all "
            "version range (`*`).",
        ),
        option(
            "--source",
            description="Specify the source from which the package should be installed.",
            flag=False,
        ),
        option(
            "--upgrade",
            description="Upgrade dependencies that are already installed and declared in the project. If the "
            "dependency is already declared, the --source option can be skipped as it will be inherited from "
            "the declaration.",
        ),
        python_option,
    ]

    def load_configuration(self, app: Application) -> None:
        return None

    def activate(self, app: Application, config: None) -> None:
        self.app = app
        app.cleo.add(self)

    def handle(self) -> int:
        from nr.util.stream import Stream

        from slap.install.installer import InstallOptions, PipInstaller, get_indexes_for_projects
        from slap.python.environment import PythonEnvironment

        if not self._validate_options():
            return 1

        result = super().handle()
        if result != 0:
            return result

        project = self.app.main_project()
        if not project or not project.is_python_project:
            self.line_error("error: not situated in a Python project", "error")
            return 1

        dependencies: dict[str, PypiDependency] = {}
        for package in self.argument("packages"):
            dependency = PypiDependency.parse(package)
            if dependency.name in dependencies:
                self.line_error(f"error: package specified more than once: <b>{dependency.name}</b>", "error")
                return 1
            dependency.source = self.option("source")
            dependencies[dependency.name] = dependency

        python = PythonEnvironment.of(get_active_python_bin(self))
        distributions = python.get_distributions(dependencies.keys())
        where = "dev" if self.option("dev") else (self.option("extra") or "run")

        to_install = (
            Stream(dependencies.values())
            .map(lambda dep: (dep, distributions[dep.name]))
            .filter(
                lambda item: self.option("upgrade") or item[1] is None or not item[0].version.accepts(item[1].version)
            )
            .map(lambda item: item[0])
            .collect()
        )

        if to_install:
            indexes = get_indexes_for_projects([project])
            config = InstallOptions(indexes, True, self.option("upgrade"))
            installer = PipInstaller(symlink_helper=None)

            self.line("Installing " + " ".join(f"<fg=cyan>{p}</fg>" for p in to_install))

            status_code = installer.install(to_install, python, config)
            if status_code != 0:
                return status_code

        distributions.update(python.get_distributions({k for k in distributions if distributions[k] is None}))
        for dep_name, dependency in dependencies.items():
            dist = distributions[dep_name]
            if not dist:
                self.line_error(
                    f"error: unable to find distribution <fg=cyan>{dep_name!r}</fg> in <s>{python.executable}</s>",
                    "error",
                )
                return 1
            if not dependency.version:
                dependency.version = VersionSpec("^" + dist.version)
            self.line(f"Adding <fg=cyan>{dependency}</fg>")
            project.add_dependency(dependency, where)

        return 0

    def _validate_options(self) -> bool:
        if not self.option("no-install") and not venv_check(self):
            return False
        if self.option("no-install") and self.option("upgrade"):
            self.line_error("error: cannot --no-install and --upgrade", "error")
            return False
        if self.option("dev") and self.option("extra"):
            self.line_error("error: cannot combine --dev and --extra", "error")
            return False
        return True



""" Represents a mutable TOML configuration file in memory. """

import typing as t
from pathlib import Path

from nr.util.generic import T


class TomlFile(t.MutableMapping[str, t.Any]):
    def __init__(self, path: Path, data: dict[str, t.Any] | None = None) -> None:
        self._path = path
        self._data: dict[str, t.Any] | None = data

    def __repr__(self) -> str:
        return f'TomlConfig(path="{self.path}")'

    def __bool__(self) -> bool:
        return self._data is not None or self.exists()

    def __len__(self) -> int:
        return len(self.load())

    def __iter__(self) -> t.Iterator[str]:
        return iter(self.load())

    def __delitem__(self, key: str) -> None:
        del self.load()[key]

    def __getitem__(self, key: str) -> t.Any:
        return self.load()[key]

    def __setitem__(self, key: str, value: t.Any) -> None:
        self.load()[key] = value

    def exists(self) -> bool:
        return self._path.is_file()

    def load(self, force_reload: bool = False) -> dict[str, t.Any]:
        import tomli

        if self._data is None or force_reload:
            with self._path.open("rb") as fp:
                self._data = tomli.load(fp)
        return self._data

    def save(self) -> None:
        import tomli_w

        if self._data is None:
            raise RuntimeError("TomlDocument is empty, call load() or value(data)")
        with self._path.open("wb") as fp:
            tomli_w.dump(self._data, fp)

    @t.overload
    def value(self) -> dict[str, t.Any]:
        ...

    @t.overload
    def value(self, data: dict[str, t.Any]) -> None:
        ...

    def value(self, data: dict[str, t.Any] | None = None) -> dict[str, t.Any] | None:
        if data is None:
            return self.load()
        else:
            self._data = data
            return None

    def value_or(self, default: T) -> dict[str, t.Any] | T:
        if self._data is not None:
            return self._data
        if self.exists():
            return self.load()
        return default

    @property
    def path(self) -> Path:
        return self._path



import datetime
import logging
import os
import time

import requests

CACHE_FILENAME = os.path.expanduser("~/.local/slap/classifiers-cache.txt")
CACHE_TTL = 60 * 60 * 24 * 7  # 7 days
CLASSIFIERS_URL = "https://pypi.org/pypi?%3Aaction=list_classifiers"
logger = logging.getLogger(__name__)
_runtime_cache: list[str] | None = None


def get_classifiers(force_refresh: bool = False) -> list[str]:
    """
    Loads the classifiers list from PyPI. Once loaded, the classifiers are cached on disk and
    in memory. The cache on disk is valid for a maxmium of seven days. Specify the *force_refresh*
    argument to ignore any caches.
    """

    global _runtime_cache
    if not force_refresh and _runtime_cache is not None:
        return list(_runtime_cache)

    def _load_cachefile():
        global _runtime_cache
        with open(CACHE_FILENAME) as fp:
            _runtime_cache = [x.rstrip("\n") for x in fp]
        return list(_runtime_cache)

    has_cachefile = not force_refresh and os.path.isfile(CACHE_FILENAME)
    if has_cachefile and (time.time() - os.path.getmtime(CACHE_FILENAME)) < CACHE_TTL:
        return _load_cachefile()
    try:
        classifiers = requests.get(CLASSIFIERS_URL, timeout=1).text.split("\n")
    except requests.exceptions.ReadTimeout as exc:
        if has_cachefile:
            cache_timestamp = datetime.datetime.fromtimestamp(os.path.getmtime(CACHE_FILENAME))
            message = " Falling back to cached classifiers (timestamp: {})".format(cache_timestamp)
        else:
            message = " Returning no classifiers. Subsequent classifier validation might raise."
        logger.warning('Error retrieving classifiers from "%s" (%s).%s', CLASSIFIERS_URL, exc, message)
        if has_cachefile:
            return _load_cachefile()
        _runtime_cache = []
        return []
    try:
        os.makedirs(os.path.dirname(CACHE_FILENAME), exist_ok=True)
        with open(CACHE_FILENAME, "w") as fp:
            fp.writelines((x + "\n" for x in classifiers))
    except BaseException:
        logger.exception("Unable to write classifiers cache file.")

    _runtime_cache = classifiers
    return list(classifiers)



import os
import shutil
import textwrap
import typing as t
from pathlib import Path

from slap.application import IO, Application, option
from slap.ext.application.venv import VenvAwareCommand
from slap.plugins import ApplicationPlugin
from slap.repository import Repository

from .install import get_active_python_bin, python_option, venv_check


class LinkCommandPlugin(VenvAwareCommand, ApplicationPlugin):
    """
    Symlink your Python package with the help of Flit.

    This command uses <u>Flit [0]</u> to symlink the Python package you are currently
    working on into your Python environment's site-packages. This is particulary
    useful if your project is using a <u>PEP 517 [1]</u> compatible build system that does
    not support editable installs.

    When you run this command, the <u>pyproject.toml</u> will be temporarily rewritten such
    that Flit can understand it. The following ways to describe a Python project are
    currently supported be the rewriter:

    1. <u>Poetry [2]</u>

      Supported configurations:
        - <fg=cyan>version</fg>
        - <fg=cyan>plugins</fg> (aka. "entrypoints")
        - <fg=cyan>scripts</fg>

    2. <u>Flit [0]</u>

      <i>Since the <opt>link</opt> command relies on Flit, no subset of configuration neeeds to be
      explicitly supported.</i>

    <b>Example usage:</b>

      <fg=yellow>$</fg> slap link
      <fg=dark_gray>Discovered modules in /projects/my_package/src: my_package
      Extras to install for deps 'all': {{'.none'}}
      Symlinking src/my_package -> .venv/lib/python3.10/site-packages/my_package</fg>

    <b>Important notes:</b>

      This command will <b>symlink</b> your package into your Python environment; this is
      much unlike a Pip editable install which instead points to your code via a
      <code>.pth</code> file. If you install something into your environment that requires an
      older version of the package you symlinked, Pip may write into those symlinked
      files and effectively change your codebase, which could lead to potential loss
      of changes.

    <u>[0]: https://flit.readthedocs.io/en/latest/</u>
    <u>[1]: https://www.python.org/dev/peps/pep-0517/</u>
    <u>[2]: https://python-poetry.org/</u>
    """

    app: Application

    name = "link"
    help = textwrap.dedent(__doc__)
    options = VenvAwareCommand.options + [
        python_option,
        option(
            "--dump-pyproject",
            description="Dump the updated pyproject.toml and do not actually do the linking.",
        ),
    ]

    def load_configuration(self, app: Application) -> None:
        return None

    def activate(self, app: Application, config: None):
        self.app = app
        app.cleo.add(self)

    def _get_source_directory(self) -> Path:
        directory = Path.cwd()
        if (src_dir := directory / "src").is_dir():
            directory = src_dir
        return directory

    def handle(self) -> int:
        result = super().handle()
        if result != 0:
            return result

        if not venv_check(self, "refusing to link"):
            return 1

        link_repository(self.io, self.app.repository, self.option("dump-pyproject"), get_active_python_bin(self))
        return 0


def link_repository(io: IO, repository: Repository, dump_pyproject: bool = False, python: str | None = None) -> None:

    from flit.install import Installer  # type: ignore[import]
    from nr.util.fs import atomic_swap

    from slap.util.pygments import toml_highlight

    # We need to pass an absolute path to Python to make sure the scripts have an absolute shebang.
    python_bin = shutil.which(python or "python")
    if not python_bin:
        raise Exception(f"Could not find Python executable from {python_bin!r}")

    # Without this set, the installer will complain about installing as the root user. If we want to
    # have a similar check in Slap, we must do it in the install command as well, otherwise you end
    # up installing as root but then just the linking step fails.
    os.environ["FLIT_ROOT_INSTALL"] = "1"

    for project in repository.get_projects_ordered():
        if not project.is_python_project:
            continue

        packages = project.packages()
        if not packages:
            continue

        for package in packages:
            config = project.pyproject_toml.value()
            dist_name = project.dist_name() or project.directory.resolve().name
            _setup_flit_config(package.name, dist_name, config)

            if dump_pyproject:
                io.write_line(f"<fg=dark_gray># {project.pyproject_toml.path}</fg>")
                io.write_line(toml_highlight(config))
                continue

            with atomic_swap(project.pyproject_toml.path, "w", always_revert=True) as fp:
                fp.close()
                project.pyproject_toml.value(config)
                project.pyproject_toml.save()
                installer = Installer.from_ini_path(
                    project.pyproject_toml.path, python=str(Path(python_bin).absolute()), symlink=True
                )
                io.write_line(f"symlinking <info>{dist_name}</info>")
                installer.install()


def _setup_flit_config(module: str, dist_name: str, data: dict[str, t.Any]) -> None:
    """Internal. Makes sure the configuration in *data* is compatible with Flit."""

    poetry = data["tool"].get("poetry", {})
    flit = data["tool"].setdefault("flit", {})
    plugins = poetry.get("plugins", {})
    scripts = poetry.get("scripts", {})
    project = data.setdefault("project", {})

    if plugins:
        project["entry-points"] = plugins
    if scripts:
        project["scripts"] = scripts

    # TODO (@NiklasRosenstein): Do we need to support gui-scripts as well?

    project["name"] = dist_name
    project["version"] = poetry["version"]
    project["description"] = ""
    flit["module"] = {"name": module}



"""
Scraper for [SPDX][1].

  [1]: https://spdx.org/licenses/
"""

from __future__ import annotations

import dataclasses
import typing as t

import requests
from databind.core.settings import Alias


@dataclasses.dataclass
class SpdxLicense:
    reference: str
    is_deprecated_license_id: t.Annotated[bool, Alias("isDeprecatedLicenseId")]
    details_url: t.Annotated[str, Alias("detailsUrl")]
    reference_number: t.Annotated[int, Alias("referenceNumber")]
    name: str
    license_id: t.Annotated[str, Alias("licenseId")]
    see_also: t.Annotated[list[str], Alias("seeAlso")]
    is_osi_approved: t.Annotated[bool, Alias("isOsiApproved")]
    is_fsf_libre: t.Annotated[bool | None, Alias("isFsfLibre")] = None

    def get_details(self) -> SpdxLicenseDetails:
        import databind.json

        response = requests.get(self.details_url)
        response.raise_for_status()
        return databind.json.load(response.json(), SpdxLicenseDetails)


@dataclasses.dataclass
class SpdxLicenseDetails:
    name: str
    license_id: t.Annotated[str, Alias("licenseId")]
    license_text: t.Annotated[str, Alias("licenseText")]
    license_text_html: t.Annotated[str, Alias("licenseTextHtml")]
    cross_ref: t.Annotated[list[dict[str, t.Any]], Alias("crossRef")]
    see_also: t.Annotated[list[str], Alias("seeAlso")]
    standard_license_template: t.Annotated[str, Alias("standardLicenseTemplate")]
    is_osi_approved: t.Annotated[bool, Alias("isOsiApproved")]
    is_deprecated_license_id: t.Annotated[bool, Alias("isDeprecatedLicenseId")]
    license_comments: t.Annotated[str | None, Alias("licenseComments")] = None
    is_fsf_libre: t.Annotated[bool | None, Alias("isFsfLibre")] = None
    standard_license_header_html: t.Annotated[str | None, Alias("standardLicenseHeaderHtml")] = None
    standard_license_header: t.Annotated[str | None, Alias("standardLicenseHeader")] = None
    standard_license_header_template: t.Annotated[str | None, Alias("standardLicenseHeaderTemplate")] = None
    deprecated_version: t.Annotated[str | None, Alias("deprecatedVersion")] = None


def wrap_license_text(license_text: str, width: int = 79) -> str:
    lines = []
    for raw_line in license_text.split("\n"):
        line = raw_line.split(" ")
        length = sum(map(len, line)) + len(line) - 1
        if length > width:
            words: list[str] = []
            length = -1
            for word in line:
                if length + 1 + len(word) >= width:
                    lines.append(" ".join(words))
                    words = []
                    length = -1
                else:
                    words.append(word)
                    length += len(word) + 1
            if words:
                lines.append(" ".join(words))
        else:
            lines.append(" ".join(line))
    return "\n".join(lines)


def get_spdx_licenses() -> dict[str, SpdxLicense]:
    """Returns a dictionary of all SPDX licenses, keyed by the license ID."""

    import databind.json

    url = "https://raw.githubusercontent.com/spdx/license-list-data/master/json/licenses.json"
    response = requests.get(url)
    response.raise_for_status()
    licenses = databind.json.load(response.json()["licenses"], list[SpdxLicense], filename=url)
    return {line.license_id: line for line in licenses}


def get_spdx_license_details(license_id: str) -> SpdxLicenseDetails:
    """Returns the details for a single SPDX license."""

    import databind.json

    url = f"https://spdx.org/licenses/{license_id}.json"
    response = requests.get(url)
    response.raise_for_status()
    return databind.json.load(response.json(), SpdxLicenseDetails, filename=url)


if __name__ == "__main__":
    import argparse

    from termcolor import colored

    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--list", action="store_true")
    parser.add_argument("license", nargs="?")
    parser.add_argument("-w", "--wrap", type=int, default=80)
    args = parser.parse_args()
    if args.list:
        for key, value in sorted(get_spdx_licenses().items()):
            print(f'{colored(key, attrs=["bold"])}: {value.name}')
    elif args.license:
        print(wrap_license_text(get_spdx_license_details(args.license).license_text, args.wrap - 1))
    else:
        parser.error("need an argument")



import dataclasses
import typing as t

from slap.check import Check, CheckResult, check, get_checks
from slap.ext.application.changelog import get_changelog_manager
from slap.plugins import CheckPlugin
from slap.project import Project


@dataclasses.dataclass
class ChangelogValidationCheckPlugin(CheckPlugin):
    """This check plugin validates the structured changelog files, if any.

    Plugin ID: `changelog`"""

    def get_project_checks(self, project: Project) -> t.Iterable[Check]:
        return get_checks(self, project)

    @check("validate")
    def _validate_changelogs(self, project: Project) -> tuple[CheckResult, str | None, str | None]:
        import tomli
        from databind.core.converter import ConversionError

        manager = get_changelog_manager(project.repository, project)
        bad_files = []
        bad_changelogs = []
        count = 0
        for changelog in manager.all():
            count += 1
            try:
                for entry in changelog.load().entries:
                    try:
                        manager.validate_entry(entry)
                    except (ConversionError, ValueError) as exc:
                        bad_changelogs.append((changelog.path.name, str(exc), entry.id))
            except (tomli.TOMLDecodeError, ConversionError) as exc:
                bad_files.append((changelog.path.name, str(exc)))

        if not count:
            return CheckResult.SKIPPED, None, None

        return (
            Check.ERROR if bad_changelogs else Check.Result.OK,
            "Broken or invalid changelogs" if bad_changelogs else f"All {count} changelogs are valid.",
            (
                "\n".join(f"<i>{fn}</i>: {err}" for fn, err in bad_files)
                if bad_files
                else ""
                + "\n"
                + "\n".join(
                    f'<i>{fn}</i>: id=<fg=yellow>"{entry_id}"</fg>: {err}' for fn, err, entry_id in bad_changelogs
                )
                if bad_changelogs
                else ""
            ).strip()
            or None,
        )



from __future__ import annotations

import shlex
import subprocess as sp
from typing import ClassVar

from slap.application import Application, argument
from slap.ext.application.venv import VenvAwareCommand
from slap.plugins import ApplicationPlugin


class RunCommandPlugin(VenvAwareCommand, ApplicationPlugin):
    """Run a command in the current active environment. If the command name is an alias
    configured in <code>[tool.slap.run]</code>, run it instead.

    In order to pass command-line options and flags, you need to add an addition `--` to
    ensure that arguments following it are parsed as positional arguments.

    Example:

        $ slap run pytest -- -vv
    """

    name: str = "run"
    requires_venv: ClassVar[bool] = False

    arguments = [
        argument(
            "args",
            description="Command name and arguments.",
            multiple=True,
        )
    ]

    def load_configuration(self, app: Application) -> dict[str, str]:
        config = (app.main_project() or app.repository).raw_config()
        return config.get("run", {})

    def activate(self, app: Application, config: dict[str, str]) -> None:
        self.app = app
        self.config = config
        app.cleo.add(self)

    def handle(self) -> int:
        result = super().handle()
        if result != 0:
            return result
        command: list[str] = self.argument("args")
        if command[0] in self.config:
            command_string = self.config[command[0]] + " " + _join_args(command[1:])
        else:
            command_string = _join_args(command)
        return sp.call(command_string, shell=True)


def _join_args(args: list[str]) -> str:
    return " ".join(map(shlex.quote, args))



from __future__ import annotations

import dataclasses
import enum
import inspect
import typing as t

if t.TYPE_CHECKING:
    from slap.application import Application
    from slap.project import Project


class CheckResult(enum.IntEnum):
    OK = enum.auto()
    RECOMMENDATION = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()
    SKIPPED = enum.auto()


@dataclasses.dataclass
class Check:
    Result: t.ClassVar[t.Type[CheckResult]] = CheckResult
    OK: t.ClassVar[CheckResult] = CheckResult.OK
    RECOMMENDATION: t.ClassVar[CheckResult] = CheckResult.RECOMMENDATION
    WARNING: t.ClassVar[CheckResult] = CheckResult.WARNING
    ERROR: t.ClassVar[CheckResult] = CheckResult.ERROR
    SKIPPED: t.ClassVar[CheckResult] = CheckResult.SKIPPED

    name: str
    result: CheckResult
    description: str | None
    details: str | None = None


def check(
    name: str,
) -> t.Callable[
    [
        t.Callable[
            [t.Any, t.Any], CheckResult | tuple[CheckResult, str | None] | tuple[CheckResult, str | None, str | None]
        ]
    ],
    t.Callable[[t.Any, t.Any], Check],
]:
    """Decorator for methods on a #CheckPlugin subclass to mark it as a function that returns detauls for a check.
    That method should return either a #CheckResult, a tuple of a #CheckResult and a description and optionally a
    third element containing more details.

    The second argument of the decorated method _must_ be annotated with either #Project or #Application to deduce
    if this method runs on an application or project.

    Example:

    ```py
    class MyCheckPlugin(CheckPlugin):
      @check("mycheck")
      def get_mycheck(self, project: Project) -> tuple[CheckResult, str]:
        return (CheckResult.OK, 'everything is in order')
    ```

    Use the #get_checks() method to run all methods on an object decorated with this decorator.
    """

    import functools

    from slap.application import Application
    from slap.project import Project

    def decorator(f: t.Callable) -> t.Callable:
        sig = inspect.signature(f)
        subject_type = sig.parameters[list(sig.parameters)[1]].annotation
        if subject_type not in (Project, Application):
            raise ValueError(f"{f} second argument must be annotated with Project or Application, got {subject_type}")

        @functools.wraps(f)
        def wrapper(self, subject) -> Check:
            result = f(self, subject)
            if isinstance(result, CheckResult):
                return Check(name, result, None, None)
            elif isinstance(result, tuple):
                return Check(name, *result)
            elif isinstance(result, Check):
                result.name = name
                return result
            else:
                raise TypeError(f"bad return value for check {name!r}: {result!r}")

        wrapper.__check_name__ = name  # type: ignore
        wrapper.__check_type__ = subject_type  # type: ignore
        return wrapper

    return decorator


def get_checks(obj: t.Any, subject: t.Union[Application, Project]) -> t.Iterable[Check]:
    """Call all methods decorated with #check() on the members of *obj*."""

    for key in dir(obj):
        value = getattr(obj, key)
        check_type = getattr(value, "__check_type__", None)
        if check_type is type(subject):
            yield value(subject)



import dataclasses
import functools
import re
from pathlib import Path

import requests

from slap.changelog import is_url
from slap.repository import Issue, PullRequest, Repository, RepositoryHost


@functools.lru_cache()
def github_get_username_from_email(api_base_url: str, email: str) -> str:
    assert email, "no email address"
    response = requests.get(f"{api_base_url}/search/users", params={"q": email})
    response.raise_for_status()
    results = response.json()
    if not results["items"]:
        raise ValueError(f"no GitHub username found for email {email!r}")
    return results["items"][0]["login"]


@dataclasses.dataclass
class GithubRepositoryHost(RepositoryHost):

    #: The owner and repository name separated by a slash. If the repository is hosted on GitHub enterprise, the domain
    #: of the GHE instance must precede the owner and repository name by another slash (e.g. `ghe.io/owner/repo`).
    repo: str

    def _get_base_url(self) -> str:
        parts = self.repo.split("/")
        if len(parts) == 3:
            return f"https://{parts[0]}"
        else:
            return "https://github.com"

    def _get_api_url(self) -> str:
        # TODO (@NiklasRosenstein): Can we rely on GHE having an `api.` subdomain?
        return self._get_base_url().replace("https://", "https://api.").rstrip("/")

    def _get_repo_url(self) -> str:
        owner, repo = self._get_repo()
        return f"{self._get_base_url()}/{owner}/{repo}"

    def _get_repo(self) -> tuple[str, str]:
        parts = self.repo.split("/")
        return parts[-2], parts[-1]

    def _get_issue_shortform(self, issue_url: str) -> str:
        match = re.search(
            r"https?://([\w\-\.]+)/(?:|.+/)([\w\-\.\_]+)/([\w\-\.\_]+)/(?:pulls?|issues)/(\d+)", issue_url
        )
        if match:
            domain, owner, repo, issue_id = match.groups()
            if domain == "github.com" and self.repo == (owner + "/" + repo):
                return issue_id
            elif self.repo == (domain + "/" + owner + "/" + repo):
                return issue_id
            result = owner + "/" + repo + "#" + issue_id
            if domain != "github.com":
                result = domain + "/" + result
            return result
        raise ValueError(f"invalid issue URL: {issue_url!r}")

    def get_username(self, repository: Repository) -> str | None:
        vcs = repository.vcs()
        assert vcs
        email = vcs.get_author().email
        username = github_get_username_from_email(self._get_api_url(), email)
        return ("@" + username) if username else email

    def get_issue_by_reference(self, issue_reference: str) -> Issue:
        issue_reference = issue_reference.lstrip("#")
        if issue_reference.isnumeric():
            id = issue_reference
            url = f"{self._get_repo_url()}/issues/{issue_reference}"
            shortform = "#" + id
        elif is_url(issue_reference):
            url = issue_reference
            shortform = self._get_issue_shortform(issue_reference)
            if shortform.isnumeric():
                id = shortform
                shortform = "#" + id
            else:
                id = shortform
        else:
            raise ValueError(f"bad issue/pull request reference for GitHub: {issue_reference!r}")
        return Issue(id, url, shortform)

    def get_pull_request_by_reference(self, pr_reference: str) -> PullRequest:
        issue = self.get_issue_by_reference(pr_reference)
        return PullRequest(issue.id, issue.url, issue.shortform)

    @staticmethod
    def detect_repository_host(repository: Repository) -> RepositoryHost | None:
        from nr.util.git import Git

        git = Git(repository.directory)
        if not git.get_toplevel():
            return None

        remotes = git.remotes()
        for remote in remotes:
            if remote.name == "origin" and "github" in remote.fetch:
                break
        else:
            return None

        match = re.search(r"github.com[:/]([^/]+/[^/]+)?", remote.fetch)
        if not match:
            return None

        repo = match.group(1)
        if repo.endswith(".git"):
            repo = repo[:-4]

        return GithubRepositoryHost(repo)

    def comment_on_issue(self, issue_reference: str, message: str) -> None:
        raise NotImplementedError

    def create_release(self, version: str, description: str, attachments: list[Path]) -> None:
        raise NotImplementedError



# We want to be at least Python 3.6 compatible, so instead of __future__.annotations we use string
# type hints and the typing module.

import ast
import functools
import operator
import typing as t

if t.TYPE_CHECKING:
    # NOTE (@NiklasRosenstein): For when we check the file individually
    from slap.python.dependency import Dependency  # type: ignore[import]


class Pep508Environment:
    """Contains the variables for evaluating PEP 508 environment markers."""

    def __init__(
        self,
        python_version: str,
        python_full_version: str,
        os_name: str,
        sys_platform: str,
        platform_release: str,
        platform_system: str,
        platform_machine: str,
        platform_python_implementation: str,
        implementation_name: str,
        implementation_version: str,
    ) -> None:
        self.python_version = python_version
        self.python_full_version = python_full_version
        self.os_name = os_name
        self.sys_platform = sys_platform
        self.platform_release = platform_release
        self.platform_system = platform_system
        self.platform_machine = platform_machine
        self.platform_python_implementation = platform_python_implementation
        self.implementation_name = implementation_name
        self.implementation_version = implementation_version

    def __repr__(self) -> str:
        args = ", ".join([f"{k}={v!r}" for k, v in self.as_json().items()])
        return f"{type(self).__name__}({args})"

    @staticmethod
    def current() -> "Pep508Environment":
        """Returns a #Pep508Environment for the current Python interpreter."""

        import os
        import platform
        import sys

        # From PEP 508
        def format_full_version(info):
            version = "{0.major}.{0.minor}.{0.micro}".format(info)
            kind = info.releaselevel
            if kind != "final":
                version += kind[0] + str(info.serial)
            return version

        return Pep508Environment(
            python_version=".".join(platform.python_version_tuple()[:2]),
            python_full_version=platform.python_version(),
            os_name=os.name,
            sys_platform=sys.platform,
            platform_release=platform.release(),
            platform_system=platform.system(),
            platform_machine=platform.machine(),
            platform_python_implementation=platform.python_implementation(),
            implementation_name=sys.implementation.name,
            implementation_version=format_full_version(sys.implementation.version),
        )

    def as_json(self) -> t.Dict[str, str]:
        return dict(vars(self))

    def evaluate_markers(
        self, markers: str, extras: t.Optional[t.Set[str]] = None, source: t.Optional[str] = None
    ) -> bool:
        """Evaluate a PEP 508 environment marker.

        Args:
          markers: The environment markers expression that is to be evaluated.
          extras: A set of extras that the `extra` marker returns true for if compared against. If this argument
            is set to `None`, using the `extra` marker in *markers* is an error.
          source: A string that is used as the filename when parsing the *markers* with #ast.parse(). If the syntax
            is invalid, this value will be included in the error message. If not specified, falls back to `<string>`.
        """

        scope = self.as_json()

        if extras is not None:

            class ExtrasEq:
                def __repr__(self) -> str:
                    return f"ExtrasEq({extras!r})"

                def __eq__(self, other) -> bool:
                    if isinstance(other, str):
                        assert extras is not None
                        return other in extras
                    return False

            scope["extra"] = t.cast(str, ExtrasEq())

        try:
            return _eval_environment_marker_ast(ast.parse(markers, filename=source or "<string>", mode="eval"), scope)
        except (ValueError, KeyError) as exc:
            raise ValueError(f"invalid environment marker string: {markers!r}\n  hint: {exc}")


def _eval_environment_marker_ast(node: ast.AST, scope: t.Dict[str, t.Any]) -> bool:
    """Evaluates an environment marker AST using the given *scope*. This is safer than using #eval()
    to avoid arbitrary code execution."""

    if isinstance(node, ast.Expression):
        return _eval_environment_marker_ast(node.body, scope)

    if isinstance(node, ast.BoolOp):
        op, initial = {
            ast.And: (operator.and_, True),
            ast.Or: (operator.or_, False),
        }[type(node.op)]
        return functools.reduce(lambda a, b: op(a, _eval_environment_marker_ast(b, scope)), node.values, initial)

    elif isinstance(node, ast.Compare):
        if len(node.ops) != 1 or len(node.comparators) != 1:
            raise ValueError("multiple comparators are not supported in environment markers")
        op = {
            ast.Eq: operator.eq,
            ast.NotEq: operator.ne,
            ast.Lt: operator.lt,
            ast.LtE: operator.le,
            ast.Gt: operator.gt,
            ast.GtE: operator.ge,
        }[type(node.ops[0])]
        return op(
            _eval_environment_markers_ast_value(node.left, scope),
            _eval_environment_markers_ast_value(node.comparators[0], scope),
        )

    raise ValueError(f"Node of type {type(node).__name__!r} not supported in environment markers")


def _eval_environment_markers_ast_value(node: ast.expr, scope: t.Dict[str, t.Any]) -> t.Any:
    """Resolve the value of an AST expression from the given *scope*."""

    if isinstance(node, ast.Name):
        try:
            return scope[node.id]
        except KeyError:
            raise ValueError(f"Marker {node.id!r} is not available in this context")

    elif isinstance(node, ast.Constant):
        return node.value

    raise ValueError(f"Node of type {type(node).__name__!r} not supported in environment markers")


def filter_dependencies(
    dependencies: t.Iterable["Dependency"], env: Pep508Environment, extras: t.Optional[t.Set[str]]
) -> t.List["Dependency"]:
    """Filters a collection of dependencies according to their environment markers and Python requirements."""

    return [dependency for dependency in dependencies if test_dependency(dependency, env, extras)]


def test_dependency(dependency: "Dependency", env: Pep508Environment, extras: t.Optional[t.Set[str]]) -> bool:
    """Tests if the *dependency* should be included given the current environment and extras."""

    if dependency.python and not dependency.python.accepts(env.python_version):
        return False
    return not dependency.markers or env.evaluate_markers(dependency.markers, extras)


if __name__ == "__main__":
    print(Pep508Environment.current())
